from typing import List

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import MigratePostgresMeta, MigrateModelMeta, json_has_key
from spinta.backends.postgresql.helpers.migrate.name import name_changed, get_pg_removed_name
from spinta.components import Context
from spinta.types.text.components import Text
from spinta.utils.nestedstruct import get_last_attr
from spinta.utils.schema import NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, Text)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: Text, model_meta: MigrateModelMeta = None, **kwargs):
    rename = meta.rename
    handler = meta.handler
    column: sa.Column = commands.prepare(context, backend, new.prop)
    columns = old.copy()

    table_name = rename.get_table_name(table.name)

    # Affected keys, are keys that have already been processed, since
    # scalar -> text conversion is done separately as well as lang rename
    # we need to make sure that we only remove / add langs that have not yet been processed
    affected_keys = []

    # Search for json column, to see if one existed beforehand
    # Remove if from columns list, to handle it separately at the end
    json_column = None
    for col in old:
        if isinstance(col.type, JSONB):
            json_column = col
            columns.remove(json_column)
            break

    # By default, json columns are all removed, if you find it, then handle removal manually
    json_column_meta = None
    if json_column is not None and model_meta is not None:
        if json_column.name in model_meta.json_columns:
            json_column_meta = model_meta.json_columns[json_column.name]
            json_column_meta.full_remove = False

    # Add empty jsonb column, if it was not found
    if json_column_meta is None and json_column is None:
        commands.migrate(context, backend, meta, table, NA, column, **kwargs)

    # Handle scalar -> text conversion
    for item in columns.copy():
        full_name = rename.get_column_name(table.name, item.name)
        key = get_last_attr(full_name)

        if json_column is not None:
            if json_has_key(backend, json_column, table, key):
                renamed_key = get_pg_removed_name(key)
                if json_has_key(backend, json_column, table, renamed_key):
                    handler.add_action(
                        ma.RemoveJSONAttributeMigrationAction(table_name, json_column, renamed_key)
                    )
                handler.add_action(
                    ma.RenameJSONAttributeMigrationAction(table_name, json_column, key, renamed_key)
                )

        col = json_column if json_column is not None else column
        handler.add_action(
            ma.TransferColumnDataToJSONMigrationAction(table_name, col, [
                (key, item)
            ])
        )
        affected_keys.append(key)
        commands.migrate(context, backend, meta, table, item, NA, **kwargs)
        columns.remove(item)

    # Handle lang rename, remove and add
    if json_column is not None:
        if json_column_meta:
            missing_keys = set(json_column_meta.keys).symmetric_difference(set(new.langs.keys()))

            # Handle renaming of keys
            for key in missing_keys:
                if key not in affected_keys:
                    formatted_name = f'{json_column.name}@{key}'
                    new_name = rename.get_column_name(table.name, formatted_name)
                    if name_changed(formatted_name, new_name):
                        extracted_key = get_last_attr(new_name)
                        json_column_meta.add_new_key(
                            key, extracted_key
                        )
                        affected_keys.append(extracted_key)
            missing_keys -= set(affected_keys)
            for missing_key in missing_keys:
                if missing_key not in json_column_meta.keys:
                    handler.add_action(ma.AddEmptyAttributeToJSONMigrationAction(
                        table=table_name,
                        source=json_column,
                        key=missing_key
                    ))
                elif missing_key not in json_column_meta.new_keys:
                    json_column_meta.add_new_key(missing_key, get_pg_removed_name(missing_key))

    # Handle column rename, which will be run at the end
    if json_column is not None and name_changed(json_column.name, column.name):
        if json_column_meta:
            json_column_meta.new_name = column.name
        else:
            commands.migrate(context, backend, meta, table, json_column, column, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, Text)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: Text, **kwargs):
    commands.migrate(context, backend, meta, table, [old], new, **kwargs)



import pathlib
from typing import Union

from spinta.components import Config
from spinta.core.config import RawConfig, DEFAULT_CONFIG_PATH


def asbool(s: Union[str, bool]) -> bool:
    if isinstance(s, bool):
        return s
    s = s.lower()
    if s in ('true', '1', 'on', 'yes'):
        return True
    if s in ('false', '0', 'off', 'no', ''):
        return False
    raise ValueError(f"Expected a boolean value, got {s!r}.")


# Returns config path from rc or DEFAULT_CONFIG_PATH
def get_config_path(rc: RawConfig) -> pathlib.Path:
    if rc is None:
        return pathlib.Path(DEFAULT_CONFIG_PATH)

    return pathlib.Path(
        rc.get('config_path') or
        DEFAULT_CONFIG_PATH
    )

# Returns id path (used for storing client files)
def get_id_path(path: pathlib.Path) -> pathlib.Path:
    return path / 'id'


# Returns helpers path (used for storing auth helper files)
def get_helpers_path(path: pathlib.Path) -> pathlib.Path:
    return path / 'helpers'


# Returns keymap.yml path, stored in helpers folder
def get_keymap_path(path: pathlib.Path):
    return get_helpers_path(path) / 'keymap.yml'


# Returns clients folder, used for auth functionality.
def get_clients_path(path: Union[Config, pathlib.Path]):
    if isinstance(path, pathlib.Path):
        return path / 'clients'
    return path.config_path / 'clients'



import dataclasses
from typing import List, Any, Dict, Union

from spinta.backends import Backend
from spinta.components import Page, Property
from spinta.core.ufuncs import Env, Expr
from spinta.types.datatype import DataType, Object
from spinta.utils.schema import NA


class QueryPage:
    page_: Page
    size: int
    select: list
    sort: list

    def __init__(self, size = None, select = None, sort = None, page = None):
        self.size = size
        self.select = [] if select is None else select
        self.sort = [] if sort is None else sort
        # By default, we set pagination to disabled
        # It gets updated if there is pagination query, which should set to most up-to-date page
        self.page_ = Page(enabled=False) if page is None else page


class QueryParams:
    prioritize_uri: bool = False
    lang_priority: List[str] = None
    expand: Expr = None
    default_expand: bool = False
    lang: List = None
    push: bool = False


class Selected:
    # Item index in select list.
    item: int = None
    # Model property if a property is selected.
    prop: Property = None
    # A value or an Expr for further processing on selected value.
    prep: Any = NA

    def __init__(
        self,
        item: int = None,
        prop: Property = None,
        # `prop` can be Expr or any other value.
        prep: Any = NA,
    ):
        self.item = item
        self.prop = prop
        self.prep = prep

    def __repr__(self):
        return self.debug()

    def __eq__(self, other):
        if isinstance(other, Selected):
            return self.item == other.item and self.prop == other.prop and self.prep == other.prep
        return False

    def debug(self, indent: str = ''):
        prop = self.prop.place if self.prop else 'None'
        if isinstance(self.prep, Selected):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep=...)\n'
                   ) + self.prep.debug(indent + '  ')
        elif isinstance(self.prep, (tuple, list)):
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={type(self.prep).__name__}...)\n'
            ) + ''.join([
                p.debug(indent + '- ')
                if isinstance(p, Selected)
                else str(p)
                for p in self.prep
            ])
        else:
            return (
                f'{indent}Selected('
                f'item={self.item}, '
                f'prop={prop}, '
                f'prep={self.prep})\n'
            )


class BaseQueryBuilder(Env):
    backend: Backend
    page: QueryPage
    expand: List[Property] = None
    query_params: QueryParams = None
    # `resolved` is used to map which prop.place properties are already
    # resolved, usually it maps to Selected, but different DataType's can return
    # different results.
    resolved: Dict[str, Selected]
    selected: Dict[str, Selected] = None

    def init_query_params(self, params: QueryParams):
        if params is None:
            params = QueryParams()
        self.query_params = params
        self._set_expanded_properties(params)

    # By default, when expand = None we think that nothing is expanded
    # in case we allow default_expand then we set it to empty list
    # if expand is empty list, we assume all are expanded
    # ?expand() will result in [] and ?expand(name) will result in [name]
    def _set_expanded_properties(self, params: QueryParams):
        prop_expr = params.expand
        if prop_expr is None and not params.default_expand:
            self.expand = None
            return

        self.expand = []
        if prop_expr is not None:
            self.expand = self.resolve(prop_expr)


class Star:
    def __str__(self):
        return '*'


class Func:
    pass


@dataclasses.dataclass
class ReservedProperty(Func):
    dtype: DataType
    param: str


@dataclasses.dataclass
class NestedProperty(Func):
    # Used to mark object nesting
    # for example object -> datatype

    left: Union[Object]
    right: Any


@dataclasses.dataclass
class ResultProperty(Func):
    # Used when result is calculated at ResultBuilder level
    # for example: checksum()
    expr: Expr


@dataclasses.dataclass
class LiteralProperty(Func):
    # Used when returning literal value
    value: Any



from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.backends.sql.ufuncs.result.components import SqlResultBuilder


@commands.get_result_builder.register(Context, Sql)
def get_result_builder(
    context: Context,
    backend: Sql
):
    return SqlResultBuilder(context)



from typing import Any
from typing import Iterator
from typing import List

from spinta.components import Context, Model
from spinta.core.ufuncs import Expr
from spinta.datasets.components import Param
from spinta.dimensions.param.components import ParamBuilder
from spinta.dimensions.param.components import ResolvedParams
from spinta.manifests.components import Manifest
from spinta.utils.schema import NA


def _recursive_iter_params(
    context: Context,
    model: Model,
    manifest: Manifest,
    params: List[Param],
    values: ResolvedParams,
) -> Iterator[ResolvedParams]:
    if len(params) > 0:
        param = params[0]
        env = ParamBuilder(context)
        env.update(params=values if values else {}, target_param=param.name, manifest=manifest)
        for source, formula in zip(param.sources, param.formulas):
            env.update(this=source)

            resolve_formula = formula or source
            expr = env.resolve(resolve_formula)
            result: Iterator[Any] = env.execute(expr)
            result = _recurse(env, param, model, resolve_formula, result)
            if not isinstance(result, Iterator):
                result = iter([result])

            for value in result:
                new_value = {**values, param.name: value}

                yield from _recursive_iter_params(
                    context,
                    model,
                    manifest,
                    params[1:],
                    new_value,
                )
    else:
        yield values


def iterparams(context: Context, model: Model, manifest: Manifest, params: List[Param] = []) -> Iterator[ResolvedParams]:
    values = {}

    if params:
        yield from _recursive_iter_params(context, model, manifest, params, values)
    else:
        yield {}


def _recurse(env: ParamBuilder, param: Param, source: Model, expr: Expr, values: Iterator) -> Iterator:
    if not isinstance(values, Iterator):
        values = iter([values])

    for value in values:
        if value is not None:
            yield value

            if param.name in param.dependencies:
                if env.target_param not in env.params or env.params[env.target_param] != value:
                    env.params[env.target_param] = value
                    if isinstance(expr, Expr):
                        yield from _recurse(env, param, source, expr, env.resolve(expr))




import itertools
import json
import textwrap
import time
import uuid
from copy import copy
from typing import Any, Tuple
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional


import pprintpp

import requests
from requests import HTTPError

import sqlalchemy as sa
import spinta.cli.push as cli_push

from spinta import spyna
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.push.components import State, PushRow, Error
from spinta.cli.helpers.push.state import check_push_state, save_push_state, save_page_values
from spinta.cli.helpers.push.utils import get_data_checksum
from spinta.components import Context
from spinta.components import Model
from spinta.core.ufuncs import asttoexpr
from spinta.utils.json import fix_data_for_json


def _prepare_rows_for_push(rows: Iterable[PushRow]) -> Iterator[PushRow]:
    for row in rows:
        row.data['_op'] = row.op
        if row.op == 'patch':
            where = {
                'name': 'eq',
                'args': [
                    {'name': 'bind', 'args': ['_id']},
                    row.data['_id'],
                ]
            }
            row.data['_where'] = spyna.unparse(where)
        elif row.op == 'insert':
            # if _revision is passed insert action gives ManagedProperty error
            if '_revision' in row.data:
                row.data.pop('_revision')
        elif row.op == 'delete' or not row.send:
            pass
        else:
            raise NotImplementedError(row.op)
        yield row


def prepare_rows_with_errors(
    client: requests.Session,
    server: str,
    context: Context,
    rows,
    model: Model,
    table: sa.Table,
    timeout: Tuple[float, float],
    error_counter: ErrorCounter = None
) -> Iterable[ModelRow]:
    conn = context.get('push.state.conn')
    for row in rows:
        type = model.model_type()
        _id = row[table.c.id]
        checksum = row[table.c.checksum]
        data = json.loads(row[table.c.data]) if row[table.c.data] else {}

        status_code, resp = send_request(
            client,
            f'{server}/{type}/{_id}',
            "GET",
            rows=[],
            data="",
            ignore_errors=[404],
            error_counter=error_counter,
            timeout=timeout
        )

        if status_code == 200:

            # Was deleted on local server, but found on target server,
            # which means we need to delete it
            if not data:
                yield prepare_rows_for_deletion(model, _id, error=True)
            # Was inserted or updated without errors
            elif checksum == get_data_checksum(resp, model):
                conn.execute(
                    table.update().
                    where(
                        (table.c.id == _id)
                    ).
                    values(
                        revision=resp['_revision'],
                        error=False,
                        data=None
                    )
                )
                yield PushRow(
                    model,
                    {'_type': type},
                    send=False
                )
                # Need to push again
            else:
                data['_revision'] = resp['_revision']
                yield PushRow(
                    model,
                    data,
                    checksum=checksum,
                    saved=True,
                    op="patch",
                    error=True
                )

        elif status_code == 404:
            # Was deleted on both - local and target servers
            if not data:
                conn.execute(
                    table.delete().
                    where(
                        (table.c.id == _id)
                    )
                )
                yield PushRow(
                    model,
                    {'_type': type},
                    send=False
                )
            # Need to push again
            else:
                yield PushRow(
                    model,
                    data,
                    checksum=checksum,
                    saved=True,
                    op="insert",
                    error=True
                )


def prepare_rows_for_deletion(
    model: Model,
    _id: str,
    error: bool = False
):
    where = {
        'name': 'eq',
        'args': [
            {'name': 'bind', 'args': ['_id']},
            _id,
        ]
    }
    return PushRow(
        model,
        {
            '_type': model.name,
            '_where': spyna.unparse(where)
        },
        saved=True,
        op="delete",
        error=error
    )


def get_row_for_error(
    rows: List[PushRow],
    errors: Optional[List[Error]] = None,
) -> str:
    message = []
    size = len(rows)

    if errors:
        rows_by_id = {
            row.data['_id']: row
            for row in rows
            if row.data and '_id' in row.data
        }
        for error in errors:
            ctx = error.get('context') or {}
            if 'id' in ctx and ctx['id'] in rows_by_id:
                row = rows_by_id[ctx['id']]
                message += [
                    f" Model {row.model.name}, data:",
                    ' ' + pprintpp.pformat(row.data),
                ]

    if len(message) == 0 and size > 0:
        row = rows[0]
        data = pprintpp.pformat(row.data)
        message += [
            (
                f" Model {row.model.name},"
                f" items in chunk: {size},"
                f" first item in chunk:"
            ),
            data
        ]

    return '\n'.join(message)


def _push_to_remote_spinta(
    client: requests.Session,
    server: str,
    rows: Iterable[PushRow],
    chunk_size: int,
    timeout: Tuple[float, float],
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[PushRow]:
    prefix = '{"_data":['
    suffix = ']}'
    slen = len(suffix)
    chunk = prefix
    ready: List[PushRow] = []

    for row in rows:
        if error_counter:
            if error_counter.has_reached_max():
                break

        data = fix_data_for_json(row.data)
        tmp = data
        if '_page' in data:
            tmp = copy(data)
            tmp.pop('_page')
        data = json.dumps(tmp, ensure_ascii=False)

        if ready and (
            len(chunk) + len(data) + slen > chunk_size or
            row.push
        ):
            yield from _send_and_receive(
                client,
                server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
                error_counter=error_counter,
                timeout=timeout
            )
            chunk = prefix
            ready = []
        if not row.push and row.send:
            chunk += (',' if ready else '') + data
            ready.append(row)

    if ready:
        if error_counter:
            if not error_counter.has_reached_max():
                yield from _send_and_receive(
                    client,
                    server,
                    ready,
                    chunk + suffix,
                    dry_run=dry_run,
                    stop_on_error=stop_on_error,
                    error_counter=error_counter,
                    timeout=timeout
                )
        else:
            yield from _send_and_receive(
                client,
                server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
                error_counter=error_counter,
                timeout=timeout
            )


def _send_and_receive(
    client: requests.Session,
    server: str,
    rows: List[PushRow],
    data: str,
    timeout: Tuple[float, float],
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[PushRow]:
    if dry_run:
        recv = _send_data_dry_run(data)
    else:
        _, recv = send_request(
            client,
            server,
            "POST",
            rows,
            data,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
            timeout = timeout
        )
        if recv:
            recv = recv['_data']
    yield from _map_sent_and_recv(rows, recv)


def _send_data_dry_run(
    data: str,
) -> Optional[List[Dict[str, Any]]]:
    """Pretend data has been sent to a target location."""
    recv = json.loads(data)['_data']
    for row in recv:
        # Extract id from delete case, it is stored in _where eq expression
        if '_op' in row and row['_op'] == 'delete' and '_where' in row:
            id_ = spyna.parse(asttoexpr(row['_where']))['args'][1]
            row['_id'] = id_

        if '_id' not in row:
            row['_id'] = str(uuid.uuid4())
        if '_revision' not in row:
            row['_revision'] = str(uuid.uuid4())
    return recv


def _map_sent_and_recv(
    sent: List[PushRow],
    recv: Optional[List[Dict[str, Any]]],
) -> Iterator[PushRow]:
    if recv is None:
        # We don't have a response, because there was an error while
        # communicating with the target server.
        for row in sent:
            row.error = True
            yield row
    else:
        assert len(sent) == len(recv), (
            f"len(sent) = {len(sent)}, len(received) = {len(recv)}"
        )
        for sent_row, recv_row in zip(sent, recv):
            if '_id' in sent_row.data:
                _id = sent_row.data['_id']
            else:
                # after delete
                _id = spyna.parse(sent_row.data['_where'])['args'][1]

            assert _id == recv_row['_id'], (
                f"sent._id = {_id}, "
                f"received._id = {recv_row['_id']}"
            )
            sent_row.data['_revision'] = recv_row['_revision']
            sent_row.error = False
            yield sent_row


def _push_rows(
    rows: Iterator[PushRow],
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> None:
    while True:
        try:
            next(rows)
        except StopIteration:
            break
        except:
            if error_counter:
                error_counter.increase()

            if stop_on_error:
                raise
            cli_push.log.exception("Error while reading data.")
        if error_counter and error_counter.has_reached_max():
            break


def send_request(
    client: requests.Session,
    server: str,
    method: str,
    rows: List[PushRow],
    data: str,
    timeout: Tuple[float, float],
    *,
    stop_on_error: bool = False,
    ignore_errors: Optional[List[int]] = None,
    error_counter: ErrorCounter = None,
) -> Tuple[Optional[int], Optional[Dict[str, Any]]]:
    data = data.encode('utf-8')
    if not ignore_errors:
        ignore_errors = []

    try:
        resp = client.request(method, server, data=data, timeout=timeout)
    except requests.exceptions.ReadTimeout as e:
        if error_counter:
            error_counter.increase()
        cli_push.log.error(
                f"Read timeout occurred. Consider using a smaller --chunk-size to avoid timeouts. Current timeout settings are (connect: {timeout[0]}s, read: {timeout[1]}s)."
        )
        if stop_on_error:
            raise
        return None, None
    except requests.exceptions.ConnectTimeout as e:
        if error_counter:
            error_counter.increase()
        cli_push.log.error(
                f"Connect timeout occurred. Current timeout settings are (connect: {timeout[0]}s, read: {timeout[1]}s)."
        )
        if stop_on_error:
            raise
        return None, None
    except IOError as e:
        if error_counter:
            error_counter.increase()
        cli_push.log.error(
            (
                "Error when sending and receiving data.%s\n"
                "Error: %s"
            ),
            get_row_for_error(rows),
            e,
        )
        if stop_on_error:
            raise
        return None, None

    try:
        resp.raise_for_status()
    except HTTPError:
        if resp.status_code not in ignore_errors:
            if error_counter:
                error_counter.increase()
            try:
                recv = resp.json()
            except requests.JSONDecodeError:
                cli_push.log.error(
                    (
                        "Error when sending and receiving data.\n"
                        "Server response (status=%s):\n%s"
                    ),
                    resp.status_code,
                    textwrap.indent(resp.text, '    '),
                )
                if stop_on_error:
                    raise
            else:
                errors = recv.get('errors')
                cli_push.log.error(
                    (
                        "Error when sending and receiving data.%s\n"
                        "Server response (status=%s):\n%s"
                    ),
                    get_row_for_error(rows, errors),
                    resp.status_code,
                    textwrap.indent(pprintpp.pformat(recv), '    '),
                )
            if stop_on_error:
                raise
        return resp.status_code, None

    return resp.status_code, resp.json()


def push(
    context: Context,
    client: requests.Session,
    server: str,  # https://example.com/
    models: List[Model],
    rows: Iterable[PushRow],
    timeout: Tuple[float, float],
    *,
    state: Optional[State] = None,
    stop_time: Optional[int] = None,    # seconds
    stop_row: Optional[int] = None,     # stop aftern given number of rows
    chunk_size: Optional[int] = None,   # split into chunks of given size in bytes
    dry_run: bool = False,              # do not send or write anything
    stop_on_error: bool = False,        # raise error immediately
    error_counter: ErrorCounter = None,
) -> None:
    if stop_time:
        rows = _add_stop_time(rows, stop_time)

    if state:
        rows = check_push_state(context, rows, state.metadata)

    rows = _prepare_rows_for_push(rows)

    if stop_row:
        rows = itertools.islice(rows, stop_row)

    rows = _push_to_remote_spinta(
        client,
        server,
        rows,
        chunk_size,
        dry_run=dry_run,
        error_counter=error_counter,
        timeout=timeout
    )
    if state and not dry_run:
        rows = save_push_state(context, rows, state.metadata)

    _push_rows(rows, stop_on_error, error_counter)


def _add_stop_time(
    rows: Iterable[PushRow],
    stop: int,  # seconds
) -> Iterator[PushRow]:
    start = time.time()
    for row in rows:
        yield row
        if time.time() - start > stop:
            break



import asyncio
from typing import List
from typing import Optional

import click
from typer import Argument
from typer import Context as TyperContext
from typer import Option

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.errors import cli_error
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.cli.helpers.migrate import MigrateRename, MigrateMeta
from spinta.cli.helpers.store import load_store
from spinta.cli.helpers.store import prepare_manifest
from spinta.core.context import configure_context
from spinta.manifests.commands.manifest import has_dataset


def bootstrap(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
):
    """Initialize backends

    This will create tables and sync manifest to backends.
    """
    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True, full_load=True)

    with context:
        require_auth(context)
        commands.bootstrap(context, store.manifest)


def sync(ctx: TyperContext):
    """Sync source manifests into main manifest

    Single main manifest can be populated from multiple different backends.
    """
    context = ctx.obj
    store = load_store(context)
    commands.load(context, store.internal, into=store.manifest)
    with context:
        require_auth(context)
        coro = commands.sync(context, store.manifest)
        asyncio.run(coro)
    click.echo("Done.")


def migrate(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
    plan: bool = Option(False, '-p', '--plan', help=(
        "If added, prints SQL code instead of executing it"
    ), is_flag=True),
    rename: str = Option(None, '-r', '--rename', help=(
        "JSON file, that maps manifest node renaming (models, properties)"
    )),
    autocommit: bool = Option(False, '-a', '--autocommit', help=(
        "If added, migrate will do atomic transactions, meaning it will automatically commit after each action (use it at your own risk)"
    )),
    datasets: Optional[List[str]] = Option(None, '-d', '--datasets', help=(
        "List of datasets to migrate"
    )),
):
    """Migrate schema change to backends"""
    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True)
    manifest = store.manifest

    if datasets:
        invalid_datasets = [dataset for dataset in datasets if not has_dataset(context, manifest, dataset)]
        if invalid_datasets:
            cli_error(f"Invalid dataset(s) provided: {', '.join(invalid_datasets)}")

    migrate_meta = MigrateMeta(
        plan=plan,
        autocommit=autocommit,
        rename=MigrateRename(
            rename_src=rename
        ),
        datasets=datasets
    )
    commands.migrate(context, manifest, migrate_meta)


def freeze(ctx: TyperContext):
    """Detect schema changes and create new version

    This will read current manifest structure, compare it with a previous
    freezed version and will generate new migration version if current and last
    versions do not match.
    """
    context = ctx.obj
    # Load just store, manifests will be loaded by freeze command.
    store = load_store(context)
    with context:
        require_auth(context)
        commands.freeze(context, store.manifest)
    click.echo("Done.")






from typing import Optional

from starlette.requests import Request
from starlette.responses import Response
from spinta import commands
from spinta.accesslog import log_response
from spinta.backends.helpers import get_select_tree, get_select_prop_names
from spinta.compat import urlparams_to_expr
from spinta.manifests.internal_sql.commands.manifest import get_model_name_list
from spinta.renderer import render
from spinta.components import Context, Namespace, Action, UrlParams
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.manifests.internal_sql.helpers import get_namespace_partial_data
from spinta.types.namespace import check_if_model_has_backend_and_source, _model_matches_params


@commands.traverse_ns_models.register(Context, Namespace, InternalSQLManifest, Action)
def traverse_ns_models(
    context: Context,
    ns: Namespace,
    manifest: InternalSQLManifest,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    internal: bool = False,
    source_check: bool = False,
    loaded: bool = False,
    **kwargs
):
    models = get_model_name_list(context, manifest, loaded, namespace=ns.name, recursive=True)
    for model_name in models:
        model = commands.get_model(context, manifest, model_name)
        if not (source_check and not check_if_model_has_backend_and_source(model)):
            if _model_matches_params(context, model, action, dataset_, resource, internal):
                yield model


@commands.getall.register(Context, Namespace, Request, InternalSQLManifest)
def getall(
    context: Context,
    ns: Namespace,
    request: Request,
    manifest: InternalSQLManifest,
    *,
    action: Action,
    params: UrlParams
):
    if params.all and params.ns:
        # for model in traverse_ns_models(context, ns, action, internal=True):
        #     commands.authorize(context, action, model)
        return _get_internal_ns_content(
            context,
            request,
            ns,
            manifest,
            params,
            action,
            recursive=True
        )
    elif params.all:
        prepare_data_for_response_kwargs = {}
        for model in commands.traverse_ns_models(context, ns, manifest, action, internal=True):
            commands.authorize(context, action, model)
            select_tree = get_select_tree(context, action, params.select)
            prop_names = get_select_prop_names(
                context,
                model,
                model.properties,
                action,
                select_tree,
            )
            prepare_data_for_response_kwargs[model.model_type()] = {
                'select': select_tree,
                'prop_names': prop_names,
            }
        expr = urlparams_to_expr(params)
        rows = commands.getall(context, ns, action=action, query=expr)
        rows = (
            commands.prepare_data_for_response(
                context,
                commands.get_model(context, manifest, row['_type']),
                params.fmt,
                row,
                action=action,
                **prepare_data_for_response_kwargs[row['_type']],
            )
            for row in rows
        )
        rows = log_response(context, rows)
        return render(context, request, ns, params, rows, action=action)
    else:
        return _get_internal_ns_content(
            context,
            request,
            ns,
            manifest,
            params,
            action
        )


def _get_internal_ns_content(
    context: Context,
    request: Request,
    ns: Namespace,
    manifest: InternalSQLManifest,
    params: UrlParams,
    action: Action,
    *,
    recursive: bool = False,
) -> Response:
    parents = [parent.name for parent in ns.parents()]
    partial_data = get_namespace_partial_data(context, manifest, ns.name, parents=parents, recursive=recursive, action=action)

    data = sorted(partial_data, key=lambda x: (x['_type'] != 'ns', x['name']))
    model = commands.get_model(context, ns.manifest, '_ns')
    select = params.select or ['name', 'title', 'description']
    select_tree = get_select_tree(context, action, select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
        auth=False,
    )
    rows = (
        commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row,
            action=action,
            select=select_tree,
            prop_names=prop_names,
        )
        for row in data
    )

    rows = log_response(context, rows)

    return render(context, request, model, params, data=rows, action=action)



from typing import Any
from typing import Dict

from spinta.core.ufuncs import Env
from spinta.datasets.components import Dataset
from spinta.manifests.components import Manifest


class ParamBuilder(Env):
    manifest: Manifest
    params: dict
    target_param: str
    stack: list


class ParamLoader(Env):
    manifest: Manifest
    dataset: Dataset


ResolvedParams = Dict[str, Any]



from typing import Optional, Dict, Any

from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta.components import Context, Action, UrlParams, Model, Node
from spinta import commands
from spinta.formats.jsonlines.components import JsonLines
from spinta.types.text.components import Text
from spinta.utils.response import aiter, peek_and_stream


@commands.render.register(Context, Request, Model, JsonLines)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


@commands.render.register(Context, Request, Node, JsonLines)
def render(
    context: Context,
    request: Request,
    node: Node,
    fmt: JsonLines,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    return _render(fmt, data, status_code, headers)


def _render(fmt: JsonLines, data, status_code: int, headers: dict):
    return StreamingResponse(
        aiter(peek_and_stream(fmt(data))),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers
    )


@commands.prepare_dtype_for_response.register(Context, JsonLines, Text, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: JsonLines,
    dtype: Text,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if 'C' in value:
        value[''] = value.pop('C')

    if len(value) == 1 and select:
        for key, data in value.items():
            key = 'C' if key == '' else key
            if key not in select.keys():
                return data

    return value



from spinta import commands
from spinta.components import Context
from spinta.manifests.internal.components import InternalManifest
from spinta.manifests.yaml.helpers import yaml_config_params
from spinta.utils.path import resource_filename


@commands.configure.register(Context, InternalManifest)
def configure(context: Context, manifest: InternalManifest):
    manifest.path = resource_filename('spinta', 'manifest')
    yaml_config_params(context, manifest)



import logging
import pathlib
from typing import List
from typing import Optional

import requests
from typer import Argument
from typer import Context as TyperContext
from typer import Exit
from typer import Option
from typer import echo

from spinta import exceptions, commands
from spinta.auth import get_client_id_from_name, get_clients_path
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ensure_data_dir
from spinta.cli.helpers.errors import ErrorCounter, cli_error
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.cli.helpers.push.components import State
from spinta.cli.helpers.push.write import push as push_
from spinta.cli.helpers.push.utils import extract_dependant_nodes, load_initial_page_data
from spinta.cli.helpers.push.read import read_rows
from spinta.cli.helpers.push.state import init_push_state
from spinta.cli.helpers.push.sync import sync_push_state
from spinta.cli.helpers.store import prepare_manifest, attach_backends, attach_keymaps
from spinta.client import get_access_token
from spinta.client import get_client_credentials
from spinta.components import Action
from spinta.components import Config
from spinta.components import Mode
from spinta.core.context import configure_context
from spinta.datasets.keymaps.sync import sync_keymap
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.utils.units import tobytes
from spinta.utils.units import toseconds

log = logging.getLogger(__name__)


def push(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output data to a given location, by default outputs to stdout"
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Push only specified dataset"
    )),
    auth: str = Option(None, '-a', '--auth', help=(
        "Authorize as a client, defaults to {default_auth_client}"
    )),
    limit: int = Option(None, help=(
        "Limit number of rows read from each model"
    )),
    chunk_size: str = Option('1m', help=(
        "Push data in chunks (1b, 1k, 2m, ...), default: 1m"
    )),
    stop_time: str = Option(None, help=(
        "Stop pushing after given time (1s, 1m, 2h, ...), by default does not "
        "stops until all data is pushed"
    )),
    stop_row: int = Option(None, help=(
        "Stop after pushing n rows, by default does not stop until all data "
        "is pushed"
    )),
    state: pathlib.Path = Option(None, help=(
        "Save push state into a file, by default state is saved to "
        "{data_path}/push/{remote}.db SQLite database file"
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    dry_run: bool = Option(False, '--dry-run', help=(
        "Read data to be pushed, but do not push or write data to the "
        "destination."
    )),
    stop_on_error: bool = Option(False, '--stop-on-error', help=(
        "Exit immediately on first error."
    )),
    no_progress_bar: bool = Option(False, '--no-progress-bar', help=(
        "Skip counting total rows to improve performance."
    )),
    retry_count: int = Option(5, '--retries', help=(
        "Repeat push until this count if there are errors."
    )),
    max_error_count: int = Option(50, '--max-errors', help=(
        "If errors exceed given number, push command will be stopped."
    )),
    incremental: bool = Option(False, '-i', '--incremental', help=(
        "Do an incremental push, only pushing objects from last page."
    )),
    page: Optional[List[str]] = Option(None, '--page', help=(
        "Page value from which rows will be pushed."
    )),
    page_model: str = Option(None, '--model', help=(
        "Model of the page value."
    )),
    synchronize: bool = Option(False, '--sync', help=(
        "Synchronize push state and keymap, in {data_path}/push/{remote}.db and {data_path}/keymap.db"
    )),
    read_timeout: float = Option(300, '--read-timeout', help=(
        "Timeout for reading a response, default: 5 minutes (300s). The value is in seconds."
    )),
    connect_timeout: float = Option(5, '--connect-timeout', help=(
        "Timeout for connecting, default: 5 seconds."
    )),
):
    """Push data to external data store"""
    synchronize_keymap = synchronize
    synchronize_state = synchronize

    if chunk_size:
        chunk_size = tobytes(chunk_size)

    if stop_time:
        stop_time = toseconds(stop_time)

    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context, full_load=True)
    config: Config = context.get('config')

    if credentials:
        credsfile = pathlib.Path(credentials)
        if not credsfile.exists():
            cli_error(
                f"Credentials file {credsfile} does not exit."
            )
    else:
        credsfile = config.credentials_file
    # TODO: Read client credentials only if a Spinta URL is given.
    creds = get_client_credentials(credsfile, output)

    if not state:
        ensure_data_dir(config.data_path / 'push')
        state = config.data_path / 'push' / f'{creds.remote}.db'

    state = f'sqlite+spinta:///{state}'

    manifest = store.manifest
    if dataset and not commands.has_dataset(context, manifest, dataset):
        cli_error(
            str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset))
        )

    ns = commands.get_namespace(context, manifest, '')

    echo(f"Get access token from {creds.server}")
    token = get_access_token(creds)

    client = requests.Session()
    client.headers['Content-Type'] = 'application/json'
    client.headers['Authorization'] = f'Bearer {token}'

    override_page = {}
    if page_model and page:
        override_page = {page_model: page}

    with context:
        auth_client = auth or config.default_auth_client
        auth_client = get_client_id_from_name(get_clients_path(config), auth_client)
        require_auth(context, auth_client)

        attach_backends(context, store, manifest)
        attach_keymaps(context, store)
        error_counter = ErrorCounter(max_count=max_error_count)

        models = commands.traverse_ns_models(context, ns, manifest, Action.SEARCH, dataset_=dataset, source_check=True)
        models = sort_models_by_ref_and_base(list(models))

        if state:
            state = State(*init_push_state(state, models))
            context.attach('push.state.conn', state.engine.begin)

        # Synchronize keymaps
        with manifest.keymap as km:
            first_time = km.first_time_sync()
            if first_time:
                synchronize_keymap = True
            dependant_models = extract_dependant_nodes(context, models, not synchronize_keymap)
            sync_keymap(
                context=context,
                keymap=km,
                client=client,
                server=creds.server,
                models=dependant_models,
                error_counter=error_counter,
                no_progress_bar=no_progress_bar,
                reset_cid=synchronize_keymap,
                dry_run=dry_run,
                timeout=(connect_timeout, read_timeout),
            )

        # Synchronize push state
        if synchronize_state:
            sync_push_state(
                context=context,
                models=models,
                client=client,
                server=creds.server,
                error_counter=error_counter,
                no_progress_bar=no_progress_bar,
                metadata=state.metadata,
                timeout=(connect_timeout, read_timeout),
            )

        initial_page_data = load_initial_page_data(
            context,
            state.metadata,
            models,
            incremental,
            override_page
        )

        rows = read_rows(
            context,
            client,
            creds.server,
            models,
            state,
            limit,
            timeout=(connect_timeout, read_timeout),
            stop_on_error=stop_on_error,
            retry_count=retry_count,
            no_progress_bar=no_progress_bar,
            error_counter=error_counter,
            initial_page_data=initial_page_data
        )

        push_(
            context,
            client,
            creds.server,
            models,
            rows,
            state=state,
            stop_time=stop_time,
            stop_row=stop_row,
            chunk_size=chunk_size,
            dry_run=dry_run,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
            timeout=(connect_timeout, read_timeout),
        )

        if error_counter.has_errors():
            raise Exit(code=1)



from sqlalchemy.dialects import registry
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.dialects.sqlite import base

# We use SQLiteDialect_pysqlite, instead of SQLiteDialect, because most of the code is the same, we only want to change
# Where we get dbapi from
# In the future, if there are errors related with sqlean and pysqlite, we can fully create our own dialect

# You use `sqlite+spinta:///...` uri to access it explicitly
# Also because we set default to spinta you can also use `sqlite:///...`
class SQLiteDialect_spinta(SQLiteDialect_pysqlite):
    driver = 'spinta'
    supports_statement_cache = True

    def __init__(self, **kwargs):
        SQLiteDialect.__init__(self, **kwargs)

    @classmethod
    def dbapi(cls):
        try:
            from sqlean import dbapi2 as sqlite
        except ImportError:
            from sqlite3 import dbapi2 as sqlite
        return sqlite


registry.register(
    "sqlite.spinta",
    __name__,
    "SQLiteDialect_spinta"
)

# This changes default sqlite dialect from pysqlite to spinta
base.dialect = SQLiteDialect_spinta



from __future__ import annotations

import re

from typing import TYPE_CHECKING, Iterable

from spinta import exceptions
from spinta.components import Config
from spinta.components import Context
from spinta.components import Model
from spinta.components import Property
from spinta.exceptions import BackendNotFound
from spinta.exceptions import InvalidName

if TYPE_CHECKING:
    from spinta.types.datatype import DataType

upper_camel_case = re.compile(r'^[A-Z][A-Za-z0-9]*$')
snake_case = re.compile(r'^[a-z](_?[a-z0-9]+)*_?$')

RESERVED_PROPERTY_NAMES = {
    '_op',
    '_type',
    '_id',
    '_revision',
    '_txn',
    '_cid',
    '_created',
    '_where',
    '_base',
    '_uri',
    '_page',
}


def check_no_extra_keys(dtype: DataType, schema: Iterable, data: Iterable):
    unknown = set(data) - set(schema)
    if unknown:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(
                dtype.prop,
                property=f'{dtype.prop.place}.{prop}',
            )
            for prop in sorted(unknown)
        )


def set_dtype_backend(dtype: DataType):
    if dtype.backend:
        backends = dtype.prop.model.manifest.store.backends
        if isinstance(dtype.backend, str):
            if dtype.backend not in backends:
                raise BackendNotFound(dtype, name=dtype.backend)
            dtype.backend = backends[dtype.backend]
    else:
        dtype.backend = dtype.prop.model.backend


def check_model_name(context: Context, model: Model):
    config: Config = context.get('config')
    if config.check_names:
        if '/' in model.name:
            name = model.name.rsplit('/', 1)[1]
        else:
            name = model.name
        if name not in ['_ns', '_txn', '_schema', '_schema/Version']:
            if upper_camel_case.match(name) is None:
                raise InvalidName(model, name=name, type='model')


def check_property_name(context: Context, prop: Property):
    config: Config = context.get('config')
    if config.check_names:
        name = prop.name
        if (
            name not in RESERVED_PROPERTY_NAMES and
            snake_case.match(name) is None
        ):
            raise InvalidName(prop, name=name, type='property')



import base64
import datetime
import decimal
import pathlib
import uuid
from typing import Any, Tuple
from typing import AsyncIterator
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

import dateutil
import shapely.geometry.base
from shapely import wkt

from geoalchemy2.elements import WKTElement, WKBElement
from geoalchemy2.shape import to_shape

from spinta import commands
from spinta import exceptions
from spinta.backends.components import Backend
from spinta.backends.components import SelectTree
from spinta.backends.helpers import check_unknown_props
from spinta.backends.helpers import flat_select_to_nested
from spinta.backends.helpers import get_model_reserved_props
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import select_keys
from spinta.backends.helpers import select_model_props
from spinta.backends.helpers import select_props
from spinta.commands import gen_object_id
from spinta.commands import is_object_id
from spinta.commands import load_operator_value
from spinta.commands import prepare
from spinta.components import Action, UrlParams, page_in_data
from spinta.components import Context
from spinta.components import DataItem
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Node
from spinta.components import Property
from spinta.core.ufuncs import asttoexpr
from spinta.exceptions import ConflictingValue, RequiredProperty, LangNotDeclared, TooManyLangsGiven, \
    UnableToDetermineRequiredLang, CoordinatesOutOfRange, InheritPropertyValueMissmatch, SRIDNotSetForGeometry, \
    InvalidUuidValue, DirectRefValueUnassignment
from spinta.exceptions import NoItemRevision
from spinta.formats.components import Format
from spinta.manifests.components import Manifest
from spinta.types.datatype import Array, ExternalRef, Inherit, PageType, BackRef, ArrayBackRef, Integer, Boolean
from spinta.types.datatype import Binary
from spinta.types.datatype import DataType
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.datatype import File
from spinta.types.datatype import JSON
from spinta.types.datatype import Number
from spinta.types.datatype import Object
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import Time
from spinta.types.datatype import UUID
from spinta.types.geometry.components import Geometry
from spinta.types.geometry.helpers import get_crs_bounding_area
from spinta.types.text.components import Text
from spinta.utils.config import asbool
from spinta.utils.encoding import encode_page_values
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable


@commands.prepare_for_write.register(Context, Model, Backend, dict)
def prepare_for_write(
    context: Context,
    model: Model,
    backend: Backend,
    patch: Dict[str, Any],
    *,
    action: Action,
    params: UrlParams = None
) -> dict:
    # prepares model's data for storing in a backend
    backend = model.backend
    result = {}
    for name, value in patch.items():
        if not name.startswith('_'):
            prop = model.properties[name]
            if params:
                value = commands.prepare_for_write(context, prop.dtype, backend, value, params)
            else:
                value = commands.prepare_for_write(context, prop.dtype, backend, value)
        result[name] = value
    return result


@commands.prepare_for_write.register(Context, Property, Backend, object)
def prepare_for_write(
    context: Context,
    prop: Property,
    backend: Backend,
    value: Any,
    *,
    action: Action,
    params: UrlParams = None
) -> Any:
    if prop.name in value:
        value[prop.name] = commands.prepare_for_write(
            context,
            prop.dtype,
            prop.dtype.backend,
            value[prop.name],
            params
        )
    return value


@commands.prepare_for_write.register(Context, DataType, Backend, object, UrlParams)
def prepare_for_write(
    context: Context,
    dtype: DataType,
    backend: Backend,
    value: Any,
    params: UrlParams
) -> Any:
    executor = commands.prepare_for_write[Context, type(dtype), type(backend), type(value)]
    result = executor(context, dtype, backend, value)
    return result


@commands.prepare_for_write.register(Context, DataType, Backend, object)
def prepare_for_write(
    context: Context,
    dtype: DataType,
    backend: Backend,
    value: Any,
) -> Any:
    # prepares value for data store
    # for simple types - loaded native values should work
    # otherwise - override for this command if necessary
    return value


@commands.prepare_for_write.register(Context, Array, Backend, list)
def prepare_for_write(
    context: Context,
    dtype: Array,
    backend: Backend,
    value: List[Any],
) -> List[Any]:
    # prepare array and it's items for datastore
    return [
        commands.prepare_for_write(context, dtype.items.dtype, backend, v)
        for v in value
    ]


@commands.prepare_for_write.register(Context, ArrayBackRef, Backend, list)
def prepare_for_write(
    context: Context,
    dtype: ArrayBackRef,
    backend: Backend,
    value: List[Any],
) -> List[Any]:
    return [
        commands.prepare_for_write(context, dtype.items.dtype, backend, v)
        for v in value
    ]


@commands.prepare_for_write.register(Context, Object, Backend, dict)
def prepare_for_write(
    context: Context,
    dtype: Object,
    backend: Backend,
    value: Dict[str, Any],
) -> Dict[str, Any]:
    # prepare objects and it's properties for datastore
    prepped = {}
    for k, v in value.items():
        prop = dtype.properties[k]
        prepped[k] = commands.prepare_for_write(context, prop.dtype, backend, v)
    return prepped


@commands.prepare_for_write.register(Context, Text, Backend, str, UrlParams)
def prepare_for_write(
    context: Context,
    dtype: Text,
    backend: Backend,
    value: str,
    params: UrlParams
) -> Any:
    default_langs = context.get('config').languages
    preferred_lang = None
    # First Step: Check Content-Language if lang exists in the property
    if params.content_langs:
        if len(params.content_langs) > 1:
            raise TooManyLangsGiven(dtype, amount=len(params.content_langs))

        if params.content_langs[0] not in dtype.langs:
            raise LangNotDeclared(dtype, lang=params.content_langs[0])

        preferred_lang = dtype.langs[params.content_langs[0]]

    # Second Step: if Content-Language not given check default languages
    if not preferred_lang:
        for lang in default_langs:
            if lang in dtype.langs:
                preferred_lang = dtype.langs[lang]
                break

    # Third Step: if Content-Language and default language does not exist, check for C language (unknown)
    if not preferred_lang:
        if not commands.identifiable(dtype.prop) and 'C' in dtype.langs:
            preferred_lang = dtype.langs['C']

    # Fourth Step: if nothing works raise Exception that language was not determined
    if not preferred_lang:
        raise UnableToDetermineRequiredLang()

    value = {preferred_lang.name: value}
    executor = commands.prepare_for_write[Context, Text, type(backend), dict]
    result = executor(context, dtype, backend, value)
    return result


@commands.prepare_for_write.register(Context, Text, Backend, dict, UrlParams)
def prepare_for_write(
    context: Context,
    dtype: Text,
    backend: Backend,
    value: dict,
    params: UrlParams
) -> Any:
    if '' in value:
        value['C'] = value.pop('')
    return value


@prepare.register(Context, Backend, Property)
def prepare(context: Context, backend: Backend, prop: Property, **kwargs):
    return prepare(context, backend, prop.dtype, **kwargs)


@commands.simple_data_check.register(Context, DataItem, Model, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    simple_model_properties_check(context, data, model, backend)


@commands.simple_data_check.register(Context, DataItem, Property, Backend)
def simple_data_check(
    context: Context,
    data: DataItem,
    prop: Property,
    backend: Backend,
) -> None:
    simple_data_check(
        context,
        data,
        prop.dtype,
        prop,
        backend,
        data.given[prop.name],
    )


def simple_model_properties_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    for name, prop in model.properties.items():
        simple_data_check(
            context,
            data,
            prop.dtype,
            prop,
            prop.dtype.backend,
            data.given.get(name, NA),
        )


@commands.simple_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    if data.action in (
        Action.UPDATE,
        Action.INSERT,
        Action.PATCH,
        Action.UPSERT
    ):
        check_type_value(dtype, value, data.action)

    # Action.DELETE is ignore for qvarn compatibility reasons.
    # XXX: make `spinta` check for revision on Action.DELETE,
    #      implementers can override this command as they please.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(prop)


@commands.simple_data_check.register(Context, DataItem, UUID, Property, Backend, str)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: UUID,
    prop: Property,
    backend: Backend,
    value: str,
) -> None:
    if not isinstance(value, uuid.UUID):
        try:
            uuid.UUID(str(value))
        except ValueError:
            raise InvalidUuidValue(prop, value=value)


@commands.simple_data_check.register(Context, DataItem, Object, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Object,
    prop: Property,
    backend: Backend,
    value: Dict[str, Any],
) -> None:
    for prop in dtype.properties.values():
        v = value.get(prop.name, NA)
        simple_data_check(context, data, prop.dtype, prop, prop.dtype.backend, v)


@commands.simple_data_check.register(Context, DataItem, Object, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Object,
    prop: Property,
    backend: Backend,
    value: Dict[str, Any],
) -> None:
    for prop in dtype.properties.values():
        v = value.get(prop.name, NA)
        simple_data_check(context, data, prop.dtype, prop, prop.dtype.backend, v)


@commands.simple_data_check.register(Context, DataItem, Text, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Text,
    prop: Property,
    backend: Backend,
    value: dict,
) -> None:
    langs = dtype.langs
    for lang, val in value.items():
        if lang not in langs:
            if lang == '' and 'C' in langs:
                continue
            raise LangNotDeclared(dtype, lang=lang)


@commands.simple_data_check.register(Context, DataItem, ExternalRef, Property, Backend, dict)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: ExternalRef,
    prop: Property,
    backend: Backend,
    value: dict,
) -> None:
    if value is None:
        return

    denorm_prop_keys = [
        key for key in dtype.properties.keys()
    ]

    if dtype.model.given.pkeys or dtype.explicit:
        allowed_keys = [prop.name for prop in dtype.refprops]
    else:
        allowed_keys = ['_id']

    allowed_keys.extend(denorm_prop_keys)
    for key in value.keys():
        if key not in allowed_keys:
            raise exceptions.FieldNotInResource(prop, property=key)
        elif key in denorm_prop_keys:
            commands.simple_data_check(
                context,
                data,
                dtype.properties[key].dtype,
                dtype.properties[key],
                backend,
                value[key]
            )

    if value and isinstance(value, dict):
        if all(val is None for key, val in value.items() if key in allowed_keys):
            raise DirectRefValueUnassignment(dtype)


@commands.simple_data_check.register(Context, DataItem, Ref, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Ref,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    if value is None:
        return

    if value and not isinstance(value, dict):
        raise exceptions.InvalidRefValue(prop, value=value)

    if isinstance(value, dict) and '_id' in value and value['_id'] is None:
        raise DirectRefValueUnassignment(dtype)


@commands.simple_data_check.register(Context, DataItem, BackRef, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: BackRef,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    if value is not NA:
        raise exceptions.CannotModifyBackRefProp(dtype)
    

@commands.simple_data_check.register(Context, DataItem, ArrayBackRef, Property, Backend, object)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: ArrayBackRef,
    prop: Property,
    backend: Backend,
    value: object,
) -> None:
    if value is not NA:
        raise exceptions.CannotModifyBackRefProp(dtype)


@commands.simple_data_check.register(Context, DataItem, Geometry, Property, Backend, str)
def simple_data_check(
    context: Context,
    data: DataItem,
    dtype: Geometry,
    prop: Property,
    backend: Backend,
    value: str,
) -> None:
    if value:
        if ';' in value:
            shape = shapely.wkt.loads(value.split(';', 1)[1])
        else:
            shape = shapely.wkt.loads(value)

        srid = dtype.srid

        if srid is None:
            raise SRIDNotSetForGeometry(dtype)

        bounding_area = get_crs_bounding_area(srid)

        if not bounding_area.contains(shape):
            raise CoordinatesOutOfRange(dtype, given=value, srid=srid, bounds=bounding_area.bounds)


@commands.complex_data_check.register(Context, DataItem, Model, Backend)
def complex_data_check(
    context: Context,
    data: DataItem,
    model: Model,
    backend: Backend,
) -> None:
    # XXX: Maybe Action.DELETE should also provide revision.
    updating = data.action in (Action.UPDATE, Action.PATCH)
    if updating and '_revision' not in data.given:
        raise NoItemRevision(model)

    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    model.properties[k],
                    given=data.given[k],
                    expected=data.saved[k],
                )

    complex_model_properties_check(context, model, backend, data)


def complex_model_properties_check(
    context: Context,
    model: Model,
    backend: Backend,
    data: DataItem,
) -> None:
    base_data = {}
    if model.base and any(isinstance(prop.dtype, Inherit) for prop in model.properties.values()):
        id_ = data.given.get('_id', None)
        if id_ is not None:
            where = {
                'name': 'eq',
                'args': [
                    {'name': 'bind', 'args': ['_id']},
                    id_,
                ]
            }
            query = asttoexpr(where)
            result = commands.getall(context, model.base.parent, backend, query=query)
            for row in result:
                base_data = row
                break

    for name, prop in model.properties.items():
        # For datasets, property type is optional.
        # XXX: But I think, it should be mandatory.
        if prop.dtype is not None:
            given = data.given.get(name, NA)

            if prop.dtype.unique and given is not NA:
                commands.check_unique_constraint(
                    context,
                    data,
                    prop.dtype,
                    prop,
                    data.backend,
                    given,
                )
            complex_data_check(
                context,
                data,
                prop.dtype,
                prop,
                data.backend,
                given,
                base_data=base_data
            )


@complex_data_check.register(Context, DataItem, DataType, Property, Backend, object)
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: DataType,
    prop: Property,
    backend: Backend,
    value: object,
    **kwargs
):
    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    dtype,
                    given=data.given[k],
                    expected=data.saved[k],
                )


@complex_data_check.register(Context, DataItem, Inherit, Property, Backend, object)
def complex_data_check(
    context: Context,
    data: DataItem,
    dtype: Inherit,
    prop: Property,
    backend: Backend,
    value: object,
    base_data: dict = {},
    **kwargs
):
    if data.action in (Action.UPDATE, Action.PATCH, Action.DELETE):
        for k in ('_type', '_revision'):
            if k in data.given and data.saved[k] != data.given[k]:
                raise ConflictingValue(
                    dtype,
                    given=data.given[k],
                    expected=data.saved[k],
                )

    if value is not NA and dtype.prop.name in base_data:
        inherited_value = base_data[dtype.prop.name]
        if inherited_value != value:
            raise InheritPropertyValueMissmatch(dtype, expected=inherited_value, given=value)


def check_type_value(dtype: DataType, value: object, action: Action):
    if dtype.required and (
        (action == Action.PATCH and value is None) or
        (action != Action.PATCH and (value is None or value is NA))
    ):
        raise RequiredProperty(dtype.prop)


@gen_object_id.register(Context, Backend, Node)
def gen_object_id(context: Context, backend: Backend, model: Node):
    return str(uuid.uuid4())


@gen_object_id.register(Context, Backend)
def gen_object_id(context: Context, backend: Backend):
    return str(uuid.uuid4())


@is_object_id.register(Context, str)
def is_object_id(context: Context, value: str):
    # Collect all available backend/model combinations.
    # TODO: this probably should be cached at load time to increase performance
    #       of object id lookup.
    # XXX: other option would be to implement two pass URL parsing, to get
    #      information about model, and then call is_object_id directly with
    #      backend and model. That would be nice.
    store = context.get('store')
    candidates = set()
    manifest = store.manifest
    for model in manifest.objects.get('model', {}).values():
        candidates.add((model.backend.__class__, model.__class__))

    # Currently all models use UUID, but dataset models use id generated from
    # other properties that form an unique id.
    for Backend_, Model_ in candidates:
        backend = Backend_()
        model = Model_()
        model.name = ''
        if is_object_id(context, backend, model, value):
            return True


@is_object_id.register(Context, Backend, Model, str)
def is_object_id(context: Context, backend: Backend, model: Model, value: str):
    try:
        return uuid.UUID(value).version == 4
    except ValueError:
        return False


@is_object_id.register(Context, Backend, Model, object)
def is_object_id(context: Context, backend: Backend, model: Model, value: object):
    # returns false if object id is non string.
    # strings are handled by other command
    return False


@is_object_id.register(Context, type(None), Namespace, object)
def is_object_id(context: Context, backend: type(None), model: Namespace, value: object):
    # Namespaces do not have object id.
    return False


@commands.prepare_data_for_response.register(Context, Namespace, Format, dict)
def prepare_data_for_response(
    context: Context,
    ns: Namespace,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_model_props(
            commands.get_model(context, ns.manifest, '_ns'),
            prop_names,
            value,
            select,
            reserved=['_id'],
        )
    }


@commands.prepare_data_for_response.register(Context, Model, Format, dict)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_model_props(
            model,
            prop_names,
            value,
            select,
            get_model_reserved_props(action, page_in_data(value)),
        )
    }


@commands.prepare_data_for_response.register(Context, Object, Format, dict)
def prepare_data_for_response(
    context: Context,
    dtype: Object,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: List[str] = None,
) -> dict:
    select = flat_select_to_nested(select)
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            value,
            data=value,
            action=action,
            select=select,
        )
        for prop, value, select in _select_prop_props(
            dtype,
            value,
            select,
            ['_type', '_revision'],
        )
    }


@commands.prepare_data_for_response.register(Context, File, Format, dict)
def prepare_data_for_response(
    context: Context,
    dtype: File,
    fmt: Format,
    value: dict,
    *,
    action: Action,
    select: List[str] = None,
    property_: Property = None,
) -> dict:
    prop = dtype.prop
    reserved = ['_type', '_revision']

    if prop.name in value and value[prop.name]:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))

    if prop.name in value and value[prop.name]:
        data = commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            value[prop.name],
            data=value,
            action=action,
            select=select,
        )
    else:
        data = {}

    for key, val, sel in select_keys(reserved, value, select):
        data[key] = val

    return data


def _select_prop_props(
    dtype: Object,
    value: dict,
    select: SelectTree,
    reserved: List[str],
):
    prop = dtype.prop
    if prop.name in value:
        check_unknown_props(prop, select, set(reserved) | set(value[prop.name]))
    else:
        check_unknown_props(prop, select, set(reserved))
    yield from select_props(
        prop,
        reserved,
        prop.model.properties,
        value,
        select,
    )
    if prop.name in value:
        yield from select_props(
            prop,
            value[prop.name].keys(),
            dtype.properties,
            value[prop.name],
            select,
        )

@commands.prepare_dtype_for_response.register(Context, Format, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    assert isinstance(value, (str, int, float, bool, type(None), list, dict)), (
        f"prepare_dtype_for_response must return only primitive, json "
        f"serializable types, {type(value)} is not a primitive data type, "
        f"model={dtype.prop.model!r}, dtype={dtype!r}"
    )
    return value


@commands.prepare_dtype_for_response.register(Context, Format, DataType, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return dtype.default


@commands.prepare_dtype_for_response.register(Context, Format, UUID, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: UUID,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return str(value)

@commands.prepare_dtype_for_response.register(Context, Format, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return {
        '_id': None,
        '_content_type': None,
    }


@commands.prepare_dtype_for_response.register(Context, Format, File, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: File,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    data = {
        key: val
        for key, val, sel in select_keys(
            ['_id', '_content_type'],
            value,
            select,
        )
    }

    if isinstance(data.get('_id'), pathlib.Path):
        # _id on FileSystem backend is a Path.
        data['_id'] = str(data['_id'])

    # File content is returned only if explicitly requested.
    if select and '_content' in select:
        if '_content' in value:
            content = value['_content']
        else:
            prop = dtype.prop
            content = commands.getfile(context, prop, dtype, dtype.backend, data=value)

        if content is None:
            data['_content'] = None
        else:
            data['_content'] = base64.b64encode(content).decode()

    return data


@commands.prepare_dtype_for_response.register(Context, Format, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Date,
    value: datetime.date,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if isinstance(value, datetime.datetime):
        return value.date().isoformat()
    else:
        return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Time, datetime.time)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Time,
    value: datetime.time,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value.isoformat()


@commands.prepare_dtype_for_response.register(Context, Format, Binary, bytes)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Binary,
    value: bytes,
    data: Dict[str, Any],
    *,
    action: Action,
    select: dict = None,
):
    return base64.b64encode(value).decode('ascii')


@commands.prepare_dtype_for_response.register(Context, Format, Geometry, shapely.geometry.base.BaseGeometry)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Geometry,
    value: shapely.geometry.base.BaseGeometry,
    data: Dict[str, Any],
    *,
    action: Action,
    select: dict = None,
):
    return value.wkt


@commands.prepare_dtype_for_response.register(Context, Format, PageType, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: PageType,
    value: list,
    data: Dict[str, Any],
    *,
    action: Action,
    select: dict = None,
):
    return encode_page_values(value).decode('ascii')


@commands.prepare_dtype_for_response.register(Context, Format, Ref, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Ref,
    value: Optional[Dict[str, Any]],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if value is None:
        return None

    properties = dtype.model.properties.copy()
    properties.update(dtype.properties)

    nullable = True
    if select and select != {'*': {}}:
        nullable = False
        names = get_select_prop_names(
            context,
            dtype,
            properties,
            action,
            select,
            reserved=['_id'],
        )
    else:
        names = value.keys()

    result = {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=data,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_props(
            dtype.model,
            names,
            properties,
            value,
            select,
        )
    }

    # Apply None checks at the end, since there might be nested values that are only calculated at the end
    # There are some cases where data changelog was not logging ref unassignment correctly and would store values as
    # empty string, instead of null
    # https://github.com/atviriduomenys/spinta/issues/556
    if nullable and all(val is None or val == "" for val in result.values()):
        return None

    return result


@commands.prepare_dtype_for_response.register(Context, Format, Ref, str)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Ref,
    value: str,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    # FIXME: Backend should never return references as strings! References
    #        should always be dicts.

    # This is a workaround for `"value": ""` cases, where they should have been `"value": null`
    # https://github.com/atviriduomenys/spinta/issues/556
    if value == "":
        return None

    return {'_id': value}


@commands.prepare_dtype_for_response.register(Context, Format, ExternalRef, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: ExternalRef,
    value: Optional[Dict[str, Any]],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if value is None:
        return None

    nullable = True
    if select and select != {'*': {}}:
        nullable = False
        names = get_select_prop_names(
            context,
            dtype,
            dtype.model.properties,
            action,
            select,
        )
    else:
        names = value.keys()

    result = {}
    props = dtype.model.properties.copy()
    props.update(dtype.properties)

    for prop, val, sel in select_props(
        dtype.model,
        names,
        props,
        value,
        select,
    ):
        result[prop.name] = commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel,
        )

    # Apply None checks at the end, since there might be nested values that are only calculated at the end
    if nullable and all(val is None for val in result.values()):
        return None

    return result


@commands.prepare_dtype_for_response.register(Context, Format, Object, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Object,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    names = get_select_prop_names(
        context,
        dtype,
        dtype.properties,
        action,
        select,
        # XXX: Not sure if auth=False should be here. But I'm too tiered to
        #      check why it does not work.
        auth=False,
    )
    return {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=data,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_props(
            dtype.prop,
            names,
            dtype.properties,
            value,
            select,
        )
    }


@commands.prepare_dtype_for_response.register(Context, Format, Array, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: list,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        dtype,
        fmt,
        value,
        data,
        action,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Format, Array, tuple)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: tuple,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        dtype,
        fmt,
        value,
        data,
        action,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Format, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value.isoformat()


def _prepare_array_for_response(
    context: Context,
    dtype: Array,
    fmt: Format,
    value: Iterable,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    if dtype.items:
        return [
            commands.prepare_dtype_for_response(
                context,
                fmt,
                dtype.items.dtype,
                v,
                data=data,
                action=action,
                select=select,
            )
            for v in value
        ]
    else:
        return [v for v in value]


@commands.prepare_dtype_for_response.register(Context, Format, Array, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Array,
    value: type(None),
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return []


@commands.prepare_dtype_for_response.register(Context, Format, JSON, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: JSON,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return value


@commands.prepare_dtype_for_response.register(Context, Format, JSON, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: JSON,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return None


@commands.prepare_dtype_for_response.register(Context, Format, Number, decimal.Decimal)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Number,
    value: decimal.Decimal,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return float(value)


@commands.prepare_dtype_for_response.register(Context, Format, Inherit, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    base_model = get_property_base_model(dtype.prop.model, dtype.prop.name)
    if base_model:
        prop = base_model.properties[dtype.prop.name]
        return commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            value,
            data=data,
            action=action,
            select=select
        )
    return None


@commands.prepare_dtype_for_response.register(Context, Format, Inherit, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if dtype.prop.name == '_base' and value:
        data = {}
        for name in value.keys():
            base_model = get_property_base_model(dtype.prop.model, name)
            if base_model:
                data.update({
                    prop.name: commands.prepare_dtype_for_response(
                        context,
                        fmt,
                        prop.dtype,
                        val,
                        data=data,
                        action=action,
                        select=sel,
                    )
                    for prop, val, sel in select_props(
                        base_model,
                        [name],
                        base_model.properties,
                        value,
                        select,
                    )
                })
        return data
    return {}


@commands.prepare_dtype_for_response.register(Context, Format, Inherit, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Inherit,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return None


@commands.prepare_dtype_for_response.register(Context, Format, ArrayBackRef, (list, tuple))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: ArrayBackRef,
    value: (list, tuple),
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return _prepare_array_for_response(
        context,
        dtype,
        fmt,
        value,
        data,
        action,
        select,
    )


@commands.prepare_dtype_for_response.register(Context, Format, ArrayBackRef, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: ArrayBackRef,
    value: type(None),
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
) -> list:
    return []


@commands.prepare_dtype_for_response.register(Context, Format, BackRef, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: BackRef,
    value: Optional[Dict[str, Any]],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if value is None or all(val is None for val in value.values()):
        return None

    if select and select != {'*': {}}:
        names = get_select_prop_names(
            context,
            dtype,
            dtype.model.properties,
            action,
            select,
            reserved=['_id'],
        )
    else:
        names = value.keys()

    data = {
        prop.name: commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=data,
            action=action,
            select=sel,
        )
        for prop, val, sel in select_props(
            dtype.model,
            names,
            dtype.model.properties,
            value,
            select,
        )
    }

    return data


@commands.prepare_dtype_for_response.register(Context, Format, BackRef, str)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: BackRef,
    value: str,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return {'_id': value}


@commands.prepare_dtype_for_response.register(Context, Format, Text, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Format,
    dtype: Text,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    if len(value) == 1 and select is not None:
        for key, data in value.items():
            if key not in select.keys():
                return data
    return value


def get_property_base_model(model: Model, name: str):
    model = model
    base_model = None
    while model.base and model.base.parent:
        model = model.base.parent
        if name in model.properties and not isinstance(
            model.properties[name].dtype,
            Inherit
        ):
            base_model = model
            break
    return base_model


@commands.unload_backend.register(Context, Backend)
def unload_backend(context: Context, backend: Backend):
    pass


@load_operator_value.register(Context, Backend, DataType, object)
def load_operator_value(
    context: Context,
    backend: Backend,
    dtype: DataType,
    value: object,
    *,
    query_params: dict,
):
    operator = query_params['name']
    # XXX: That is probably not a very reliable way of getting original operator
    #      name. Maybe at least this should be documented some how.
    operator_name = query_params.get('origin', operator)
    if operator in ('startswith', 'contains') and not isinstance(dtype, (String, PrimaryKey)):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)
    if operator in ('gt', 'ge', 'lt', 'le') and isinstance(dtype, String):
        raise exceptions.InvalidOperandValue(dtype, operator=operator_name)


@commands.make_json_serializable.register(Model, dict)
def make_json_serializable(model: Model, value: dict) -> dict:
    return commands.make_json_serializable[Object, dict](model, value)


@commands.make_json_serializable.register(DataType, object)
def make_json_serializable(dtype: DataType, value: object):
    return value


@commands.make_json_serializable.register(DateTime, datetime.datetime)
def make_json_serializable(dtype: DateTime, value: datetime.datetime):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.date)
def make_json_serializable(dtype: Date, value: datetime.date):
    return value.isoformat()


@commands.make_json_serializable.register(Date, datetime.datetime)
def make_json_serializable(dtype: Date, value: datetime.datetime):
    return value.date().isoformat()


def _get_json_serializable_props(dtype: Object, value: dict):
    for k, v in value.items():
        if k in dtype.properties:
            yield dtype.properties[k], v
        else:
            Exception(f"Unknown {k} key in {dtype!r} object.")


@commands.make_json_serializable.register(Object, dict)
def make_json_serializable(dtype: Object, value: dict):
    return {
        prop.name: commands.make_json_serializable(prop.dtype, v)
        for prop, v in _get_json_serializable_props(dtype, value)
    }


@commands.make_json_serializable.register(Array, list)
def make_json_serializable(dtype: Array, value: list):
    return [make_json_serializable(dtype.items.dtype, v) for v in value]


@commands.update.register(Context, Property, DataType, Backend)
def update(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.patch.register(Context, Property, DataType, Backend)
def patch(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


@commands.delete.register(Context, Property, DataType, Backend)
def delete(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    dstream = _prepare_property_for_delete(prop, dstream)
    return commands.update(
        context, prop.model, prop.model.backend,
        dstream=dstream,
        stop_on_error=stop_on_error,
    )


async def _prepare_property_for_delete(
    prop: Property,
    dstream: AsyncIterator[DataItem],
) -> AsyncIterator[DataItem]:
    async for data in dstream:
        if data.patch:
            data.patch = {
                **{k: v for k, v in data.patch.items() if k.startswith('_')},
                prop.name: None,
            }
        yield data


@commands.cast_backend_to_python.register(Context, Model, Backend, dict)
def cast_backend_to_python(
    context: Context,
    model: Model,
    backend: Backend,
    data: dict,
) -> dict:
    return {
        k: commands.cast_backend_to_python(
            context,
            model.properties[k].dtype,
            backend,
            v,
        ) if k in model.properties else v
        for k, v in data.items()
    }


@commands.cast_backend_to_python.register(Context, Property, Backend, object)
def cast_backend_to_python(
    context: Context,
    prop: Property,
    backend: Backend,
    data: object,
) -> dict:
    return commands.cast_backend_to_python(context, prop.dtype, backend, data)


@commands.cast_backend_to_python.register(Context, DataType, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: DataType,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    return data


@commands.cast_backend_to_python.register(Context, UUID, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: UUID,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            return uuid.UUID(str(data))
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, DateTime, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: DateTime,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            return dateutil.parser.isoparse(data)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Time, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: Time,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str) and ":" in data:
        try:
            isoparser = dateutil.parser.isoparser()
            return isoparser.parse_isotime(data)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Date, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: Date,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            isoparser = dateutil.parser.isoparser()
            return isoparser.parse_isodate(data)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Integer, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: Integer,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            new = data
            if "," in data and "." not in data and data.count(",") == 1:
                new = data.replace(",", ".")
            elif "," in data:
                new = data.replace(",", "")
            return int(new)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Number, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: Number,
    backend: Backend,
    data: Any,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            new = data
            if "," in data and "." not in data and data.count(",") == 1:
                new = data.replace(",", ".")
            elif "," in data:
                new = data.replace(",", "")
            return float(new)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Binary, Backend, str)
def cast_backend_to_python(
    context: Context,
    dtype: Binary,
    backend: Backend,
    data: str,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            return base64.b64decode(data)
        except Exception as e:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Boolean, Backend, str)
def cast_backend_to_python(
    context: Context,
    dtype: Boolean,
    backend: Backend,
    data: str,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            return asbool(data)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Geometry, Backend, str)
def cast_backend_to_python(
    context: Context,
    dtype: Geometry,
    backend: Backend,
    data: str,
) -> Any:
    if _check_if_nan(data):
        return None
    if isinstance(data, str):
        try:
            return wkt.loads(data)
        except:
            return data
    return data


@commands.cast_backend_to_python.register(Context, Geometry, Backend, WKTElement)
def cast_backend_to_python(
    context: Context,
    dtype: Geometry,
    backend: Backend,
    data: WKTElement,
) -> Any:
    if _check_if_nan(data):
        return None
    return to_shape(data)


@commands.cast_backend_to_python.register(Context, Geometry, Backend, WKBElement)
def cast_backend_to_python(
    context: Context,
    dtype: Geometry,
    backend: Backend,
    data: WKBElement,
) -> Any:
    if _check_if_nan(data):
        return None
    return to_shape(data)


@commands.cast_backend_to_python.register(Context, Ref, Backend, object)
def cast_backend_to_python(
    context: Context,
    dtype: Ref,
    backend: Backend,
    data: Dict[str, Any],
) -> Any:
    if data:
        result = {}
        for key, value in data.items():
            converted = value
            new_type = None
            for item in dtype.refprops:
                if item.name == key:
                    new_type = item.dtype
                    break
            if new_type:
                converted = commands.cast_backend_to_python(
                    context,
                    new_type,
                    backend,
                    value
                )
            result[key] = converted

        if not result or all(value is None for value in result.values()):
            return None

        return result
    return data


@commands.cast_backend_to_python.register(Context, Object, Backend, dict)
def cast_backend_to_python(
    context: Context,
    dtype: Object,
    backend: Backend,
    data: Dict[str, Any],
) -> Dict[str, Any]:
    if data:
        return {
            k: commands.cast_backend_to_python(
                context,
                dtype.properties[k].dtype,
                backend,
                v,
            ) if k in dtype.properties else v
            for k, v in data.items()
        }
    return data


@commands.cast_backend_to_python.register(Context, Array, Backend, list)
def cast_backend_to_python(
    context: Context,
    dtype: Array,
    backend: Backend,
    data: List[Any],
) -> List[Any]:
    if data and dtype:
        return [
            commands.cast_backend_to_python(
                context,
                dtype.items.dtype,
                backend,
                v,
            )
            for v in data
        ]
    return data


@commands.reload_backend_metadata.register(Context, Manifest, Backend)
def reload_backend_metadata(context, manifest, backend):
    pass


@commands.reload_backend_metadata.register(Context, Manifest, type(None))
def reload_backend_metadata(context, manifest, backend):
    pass


def _check_if_nan(value: Any) -> bool:
    # Check for nan values, IEEE 754 defines that comparing with nan always returns false
    if value != value:
        return True
    return False


@commands.get_error_context.register(Backend)
def get_error_context(backend: Backend, *, prefix='this') -> Dict[str, str]:
    return {
        'type': f'{prefix}.type',
        'name': f'{prefix}.name',
        'origin': f'{prefix}.origin',
        'features': f'{prefix}.features',
    }



from collections import abc

from multipledispatch import dispatch
from typing import Any, Union

from spinta import commands
from spinta.backends import Backend
from spinta.components import Context, Property
from spinta.core.ufuncs import Expr
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.exceptions import ValueNotInEnum, PropertyNotFound
from spinta.ufuncs.basequerybuilder.components import Selected
from spinta.ufuncs.resultbuilder.components import ResultBuilder
from spinta.utils.schema import NA

ResultBuilderGetter = abc.Callable[[], ResultBuilder]


def _resolve_expr(context: Context, row: Any, sel: Selected, result_builder_getter: Union[ResultBuilderGetter, ResultBuilder]) -> Any:
    if sel.item is None:
        val = None
    else:
        val = row[sel.item]

    result_builder = result_builder_getter
    if isinstance(result_builder, abc.Callable):
        result_builder = result_builder()
    env = result_builder.init(val, sel.prop, row)
    return env.resolve(sel.prep)


def _aggregate_values(data: Any, target: Property):
    key_path = target.place
    key_parts = key_path.split('.')

    # Drop first part, since if nested prop is part of the list will always be first value
    # ex: from DB we get {"notes": [{"note": 0}]}
    # but after fetching the value we only get [{"note": 0}]
    # so if our place is "notes.note", we need to drop "notes" part
    if len(key_parts) > 1:
        key_parts = key_parts[1:]

    def recursive_collect(sub_data, depth=0):
        if depth < len(key_parts):
            if isinstance(sub_data, list):
                collected = []
                for item in sub_data:
                    collected.extend(recursive_collect(item, depth))
                return collected
            elif isinstance(sub_data, dict) and key_parts[depth] in sub_data:
                return recursive_collect(sub_data[key_parts[depth]], depth + 1)
        else:
            return [sub_data]

        return []

    # Start the recursive collection process
    return recursive_collect(data, 0)


# This is the fastest way to call this function, use this if you do not want to recreate `ResultBuilder`
@dispatch(Context, ResultBuilder, object, object)
def get_row_value(context: Context, result_builder: ResultBuilder, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        result_builder
    )


# This is the fastest way to call this function, use this if you do not want to recreate `ResultBuilder`
@dispatch(Context, ResultBuilder, object, object, bool)
def get_row_value(context: Context, result_builder: ResultBuilder, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        result_builder
    )


# This is the second-fastest way to call this function, it allows for `ResultBuilder` recreation
@dispatch(Context, abc.Callable, object, object, bool)
def get_row_value(context: Context, result_builder_getter: ResultBuilderGetter, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        result_builder_getter
    )


# This is the second-fastest way to call this function, it allows for `ResultBuilder` recreation
@dispatch(Context, abc.Callable, object, object)
def get_row_value(context: Context, result_builder_getter: ResultBuilderGetter, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        result_builder_getter
    )


# This is the slowest way to call this function, use this only if you cannot pass `backend_result_builder_getter` directly
@dispatch(Context, Backend, object, object)
def get_row_value(context: Context, backend: Backend, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        backend_result_builder_getter(context, backend)
    )


# This is the slowest way to call this function, use this only if you cannot pass `backend_result_builder_getter` directly
@dispatch(Context, Backend, object, object, bool)
def get_row_value(context: Context, backend: Backend, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        backend_result_builder_getter(context, backend)
    )


def backend_result_builder_getter(context: Context, backend: Backend) -> ResultBuilderGetter:
    return lambda: commands.get_result_builder(context, backend)


def _get_row_value(
    context: Context,
    row: Any,
    sel: Any,
    check_enums: bool,
    result_builder_getter: Union[ResultBuilderGetter, ResultBuilder]
):
    if isinstance(sel, Selected):
        if isinstance(sel.prep, Expr):
            val = _resolve_expr(context, row, sel, result_builder_getter)
        elif sel.prep is not NA:
            val = _get_row_value(context, row, sel.prep, check_enums, result_builder_getter)
        else:
            if sel.item is not None:
                val = row[sel.item]
                prop = sel.prop
                if prop is not None and prop.list is not None:
                    val = _aggregate_values(val, prop)
            else:
                val = None

        if check_enums:
            if enum := get_prop_enum(sel.prop):
                if val is None:
                    pass
                elif str(val) in enum:
                    item = enum[str(val)]
                    if item.prepare is not NA:
                        val = item.prepare
                else:
                    raise ValueNotInEnum(sel.prop, value=val)

        return val
    if isinstance(sel, tuple):
        return tuple(_get_row_value(context, row, v, check_enums, result_builder_getter) for v in sel)
    if isinstance(sel, list):
        return [_get_row_value(context, row, v, check_enums, result_builder_getter) for v in sel]
    if isinstance(sel, dict):
        return {k: _get_row_value(context, row, v, check_enums, result_builder_getter) for k, v in sel.items()}
    return sel



import datetime
from typer import echo
from typing import List

from spinta.auth import authorized
from spinta.cli.helpers.errors import ErrorCounter
from spinta.components import Context, Model, Action, Property, Config
from spinta.datasets.backends.helpers import extract_values_from_row
from spinta.datasets.keymaps.components import KeyMap
import tqdm

from spinta.utils.response import get_request


def _build_changelog_url(
    server: str,
    model: str,
    offset_cid: int,
    limit: int
) -> str:
    url = f'{server}/{model}/:changes'
    if offset_cid:
        url = f'{url}/{offset_cid + 1}'
    url = f'{url}?limit({limit})'
    return url


def _fetch_changelog_data(
    config: Config,
    model: Model,
    client,
    server: str,
    offset_cid: int,
    error_counter: ErrorCounter,
    timeout: tuple[float, float]
):
    limit = config.sync_page_size
    offset = offset_cid
    while True:
        url = _build_changelog_url(
            server=server,
            model=model.model_type(),
            offset_cid=offset,
            limit=limit
        )
        status_code, resp = get_request(
            client,
            url,
            error_counter=error_counter,
            timeout=timeout
        )
        if status_code != 200:
            break

        data = resp.get('_data')
        if not data:
            break

        cid = offset
        i = 0
        for i, row in enumerate(data, start=1):
            cid = row['_cid']
            yield row

        # Do not iterate further if limit was not reached
        if i < limit:
            break

        offset = cid


def sync_keymap(
    context: Context,
    keymap: KeyMap,
    client,
    server: str,
    models: List[Model],
    error_counter: ErrorCounter,
    no_progress_bar: bool,
    reset_cid: bool,
    timeout: tuple[float, float],
    dry_run: bool = False,
):
    config = context.get('config')
    counters = {}
    if not no_progress_bar:
        counters = {
            '_total': tqdm.tqdm(desc='SYNCHRONIZING KEYMAP', ascii=True)
        }
        echo()
    for model in models:
        model_keymaps = [model.model_type()]
        primary_keys = model.external.pkeys
        skip_model = False
        for key in primary_keys:
            is_authorized = authorized(context, key, action=Action.SEARCH)
            if not is_authorized:
                echo(f"SKIPPED KEYMAP '{model.model_type()}' MODEL SYNC, NO PERMISSION.")
                skip_model = True
                break
        if skip_model:
            continue

        for combination in model.required_keymap_properties:
            model_keymaps.append(f'{model.model_type()}.{"_".join(combination)}')

        # Get min cid (in case new model was added, or models are out of sync)
        sync_cid = 0
        if not reset_cid:
            initial = True
            for keymap_name in model_keymaps:
                cid = keymap.get_sync_data(keymap_name)
                if not cid:
                    sync_cid = 0
                    break

                if initial:
                    sync_cid = cid
                    initial = False
                else:
                    sync_cid = min(sync_cid, cid)

        data = _fetch_changelog_data(
            config=config,
            model=model,
            client=client,
            server=server,
            offset_cid=sync_cid,
            error_counter=error_counter,
            timeout = timeout
        )

        if not no_progress_bar:
            model_counters = {}
            for keymap_name in model_keymaps:
                model_counters[keymap_name] = tqdm.tqdm(desc=keymap_name, ascii=True)
            counters[model.model_type()] = model_counters

        for row in data:
            if row['_op'] == 'insert':
                sync_model_insert(
                    keymap=keymap,
                    model=model,
                    row=row,
                    primary_keys=primary_keys,
                    counters=counters,
                    dry_run=dry_run
                )
            elif row['_op'] == 'patch' or row['_op'] == 'update':
                sync_model_update(
                    keymap=keymap,
                    model=model,
                    row=row,
                    counters=counters,
                    dry_run=dry_run
                )
            sync_cid = row['_cid']

        if not dry_run:
            for keymap_name in model_keymaps:
                keymap.update_sync_data(keymap_name, sync_cid, datetime.datetime.now())

        if model.model_type() in counters:
            for counter in counters[model.model_type()].values():
                counter.close()

    if '_total' in counters:
        counters['_total'].close()


def remap_decoded_values(decoded: object, properties: List[str]):
    mapped = {}
    if not isinstance(decoded, list):
        decoded = [decoded]
    for i, prop in enumerate(properties):
        mapped[prop] = decoded[i]
    return mapped


def keymaps_affected_by_change(row: dict, model: Model) -> dict:
    keymaps = {}
    for key in model.external.pkeys:
        if key.name in row.keys():
            keymaps[model.model_type()] = list([prop.name for prop in model.external.pkeys])
            break

    for combination in model.required_keymap_properties:
        for prop in combination:
            if prop in row.keys():
                keymaps[f'{model.model_type()}.{"_".join(combination)}'] = combination
                break

    return keymaps


def _extract_row_data_from_keys(row: dict, keys: List[Property]) -> dict:
    result = {}
    for key in keys:
        if key.name in row:
            result[key.name] = row[key.name]
    return result


def sync_model_insert(keymap: KeyMap, model: Model, row: dict, primary_keys: List[Property], counters: dict, dry_run: bool):
    values = _extract_row_data_from_keys(row, primary_keys)
    value = list(values.values())
    if len(value) == 1:
        value = value[0]

    if not dry_run:
        keymap.synchronize(model.model_type(), value, row['_id'])

    if model.model_type() in counters and model.model_type() in counters[model.model_type()]:
        counters[model.model_type()][model.model_type()].update(1)
    if '_total' in counters:
        counters['_total'].update(1)

    if row['_id'] and model.required_keymap_properties:
        for combination in model.required_keymap_properties:
            joined = '_'.join(combination)
            key = f'{model.model_type()}.{joined}'
            val = extract_values_from_row(row, model, combination)

            if not dry_run:
                keymap.synchronize(key, val, row['_id'])

            if model.model_type() in counters and key in counters[model.model_type()]:
                counters[model.model_type()][key].update(1)
            if '_total' in counters:
                counters['_total'].update(1)


def sync_model_update(keymap: KeyMap, model: Model, row: dict, counters: dict, dry_run: bool):
    for km, props in keymaps_affected_by_change(row, model).items():
        decoded = keymap.decode(km, row['_id'])
        remapped = remap_decoded_values(decoded, props)
        for key in remapped.keys():
            if key in row:
                remapped[key] = row[key]
        val = list(remapped.values())
        if len(val) == 1:
            val = val[0]

        if not dry_run:
            keymap.synchronize(km, val, row['_id'])

        if model.model_type() in counters and km in counters[model.model_type()]:
            counters[model.model_type()][km].update(1)
        if '_total' in counters:
            counters['_total'].update(1)



from typing import List, Any, Tuple, Dict

from spinta.components import Page, Property
from spinta.core.ufuncs import ufunc, Expr, Negative, Bind, GetAttr
from spinta.datasets.backends.sql.ufuncs.components import Selected
from spinta.exceptions import InvalidArgumentInExpression, CannotSelectTextAndSpecifiedLang, \
    LangNotDeclared, FieldNotInResource
from spinta.types.datatype import DataType, String, Ref, Object, Array, File, BackRef, PrimaryKey, ExternalRef
from spinta.types.text.components import Text
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, Star, ReservedProperty, NestedProperty, \
    ResultProperty, LiteralProperty
from spinta.ufuncs.basequerybuilder.helpers import get_pagination_compare_query, process_literal_value
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.schema import NA
from spinta.utils.types import is_value_literal

# This file contains reusable resolvers, that should be backend independent
# in case there are cases where you need to have backend specific, just overload them
# keep in mind, that `select(env, expr)` is written backend specific, at least for now

# The main goal of this resolver logic, is that you have `_resolve_getattr`
# which is responsible for returning right types to process further
# Bind mean that it's a leaf node and can be accessed directly
# GetAttr mean that it can be nested further
# ForeignProperty is used, when you need to access properties from other models
# ReservedProperty is used, when datatypes do not have direct association to required field
#   example: File (when accessing child properties), Ref (when accessing it's own key, without needing to join models)


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(BaseQueryBuilder, Bind, Bind, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    field: Bind,
    attr: Bind
):
    return GetAttr(field.name, attr)


@ufunc.resolver(BaseQueryBuilder, Bind, GetAttr, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    obj: Bind,
    attr: GetAttr
):
    return GetAttr(obj.name, attr)


@ufunc.resolver(BaseQueryBuilder, GetAttr, Bind, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    obj: GetAttr,
    attr: Bind
):
    leaf = env.call('getattr', obj.name, attr)
    return GetAttr(obj.obj, leaf)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    attr: GetAttr,
) -> ForeignProperty:
    prop = env.model.properties[attr.obj]
    return env.call('_resolve_getattr', prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Ref, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Ref, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(BaseQueryBuilder, Ref, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Ref,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        return dtype.properties[attr.obj].dtype
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, Ref, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Ref,
    attr: Bind,
):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype

    # Check for self reference, no need to do joins if table already contains the value
    if attr.name == '_id':
        return ReservedProperty(dtype, attr.name)

    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, ExternalRef, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: ExternalRef,
    attr: Bind,
):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype

    # Check for self reference, no need to do joins if table already contains the value
    for refprop in dtype.refprops:
        if refprop.name == attr.name:
            return ReservedProperty(dtype, attr.name)

    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, BackRef, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: BackRef,
    attr: GetAttr,
):
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, BackRef, Bind)
def _resolve_getattr(env, dtype, attr):
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, BackRef, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, BackRef, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(BaseQueryBuilder, Text, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Text,
    bind: Bind
):
    if dtype.prop.name in env.model.properties:
        prop = env.model.properties[dtype.prop.name]
        if bind.name in prop.dtype.langs:
            return prop.dtype.langs[bind.name].dtype
        raise LangNotDeclared(dtype, lang=bind.name)


@ufunc.resolver(BaseQueryBuilder, Object, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Object,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        prop = dtype.properties[attr.obj]

        return NestedProperty(
            left=dtype,
            right=env.call('_resolve_getattr', prop.dtype, attr.name)
        )
    raise FieldNotInResource(dtype, property=attr.obj)


@ufunc.resolver(BaseQueryBuilder, Object, Bind)
def _resolve_getattr(env, dtype, attr):
    if attr.name in dtype.properties:
        return NestedProperty(
            left=dtype,
            right=dtype.properties[attr.name].dtype
        )
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(BaseQueryBuilder, Array, (Bind, GetAttr))
def _resolve_getattr(env, dtype, attr):
    return NestedProperty(
        left=dtype,
        right=env.call('_resolve_getattr', dtype.items.dtype, attr)
    )


@ufunc.resolver(BaseQueryBuilder, File, Bind)
def _resolve_getattr(env, dtype, attr):
    return ReservedProperty(dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def select(env: BaseQueryBuilder, attr: GetAttr) -> Selected:
    resolved = env.call('_resolve_getattr', attr)
    return env.call('select', resolved)


@ufunc.resolver(BaseQueryBuilder, NestedProperty)
def select(env: BaseQueryBuilder, nested: NestedProperty) -> Selected:
    return Selected(
        prop=nested.left.prop,
        prep=env.call('select', nested.right),
    )


@ufunc.resolver(BaseQueryBuilder, ReservedProperty)
def select(env, prop):
    return env.call('select', prop.dtype, prop.param)


@ufunc.resolver(BaseQueryBuilder, ResultProperty)
def select(env: BaseQueryBuilder, prop: ResultProperty):
    return Selected(
        prep=prop.expr
    )


@ufunc.resolver(BaseQueryBuilder, ForeignProperty)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Property)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external and prop.external.prepare is not NA:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, tuple)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, list)
def select(
    env: BaseQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(BaseQueryBuilder, tuple)
def select(
    env: BaseQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(BaseQueryBuilder, dict)
def select(
    env: BaseQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: Add tests.
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(BaseQueryBuilder, LiteralProperty)
def select(
    env: BaseQueryBuilder,
    prep: LiteralProperty,
) -> Selected:
    return Selected(prep=prep.value)


@ufunc.resolver(BaseQueryBuilder, DataType, object)
def select(env: BaseQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if prop.external and prop.external.name:
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def sort(env, field):
    dtype = env.call('_resolve_getattr', field)
    return env.call('sort', dtype)


@ufunc.resolver(BaseQueryBuilder, Expr, name='paginate')
def paginate(env, expr):
    if len(expr.args) != 1:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')
    page = expr.args[0]
    if isinstance(page, Page):
        if page.enabled:
            if not page.filter_only:
                env.page.select = env.call('select', page)
                for by, page_by in page.by.items():
                    sorted_ = env.call('sort',
                                       Negative(page_by.prop.name) if by.startswith("-") else Bind(page_by.prop.name))
                    if sorted_ is not None:
                        if isinstance(sorted_, (list, set, tuple)):
                            env.page.sort += sorted_
                        else:
                            env.page.sort.append(sorted_)
            env.page.page_ = page
            env.page.size = page.size
            return env.resolve(get_pagination_compare_query(page))
    else:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')


@ufunc.resolver(BaseQueryBuilder, Expr, name='expand')
def expand(env, expr):
    result = []
    if expr.args:
        for arg in expr.args:
            resolved = env.resolve(arg)
            result.append(resolved)
        return result
    return None


@ufunc.resolver(BaseQueryBuilder, str, name='op')
def op_(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(BaseQueryBuilder, Expr, name='page')
def page_(env, expr):
    pass


@ufunc.resolver(BaseQueryBuilder, DataType, list)
def validate_dtype_for_select(env, dtype: DataType, selected_props: List[Property]):
    raise NotImplemented(f"validate_dtype_for_select with {dtype.name} is not implemented")


@ufunc.resolver(BaseQueryBuilder, Text, list)
def validate_dtype_for_select(env, dtype: Text, selected_props: List[Property]):
    for prop in selected_props:
        if dtype.prop.name == prop.name or prop.parent == dtype.prop:
            raise CannotSelectTextAndSpecifiedLang(dtype)


@ufunc.resolver(BaseQueryBuilder, String, list)
def validate_dtype_for_select(env, dtype: String, selected_props: List[Property]):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        parent = dtype.prop.parent
        for prop in selected_props:
            if parent.name == prop.name or prop == dtype.prop:
                raise CannotSelectTextAndSpecifiedLang(parent.dtype)


@ufunc.resolver(BaseQueryBuilder, NestedProperty, object, names=COMPARE)
def compare(env: BaseQueryBuilder, op: str, nested: NestedProperty, value: object):
    return env.call(op, nested.right, value)


@ufunc.resolver(BaseQueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = env.model.get_from_flatprops(field.name)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(BaseQueryBuilder, Expr)
def testlist(env: BaseQueryBuilder, expr: Expr) -> tuple:
    args, kwargs = expr.resolve(env)
    result = []
    for arg in args:
        result.append(process_literal_value(arg))
    return tuple(result)



from collections.abc import Iterator
from typing import Any

from spinta import commands
from spinta.components import Model
from spinta.core.ufuncs import Bind, Expr
from spinta.core.ufuncs import Env
from spinta.core.ufuncs import ufunc
from spinta.datasets.components import Param
from spinta.dimensions.param.components import ParamBuilder, ParamLoader
from spinta.exceptions import PropertyNotFound, KeyNotFound, ModelNotFound, InvalidParamSource
from spinta.utils.schema import NotAvailable


@ufunc.resolver(Env, Bind)
def param(env: Env, bind: Bind) -> Any:
    return env.params[bind.name]


@ufunc.resolver(ParamBuilder, Model)
def read(env: ParamBuilder, model: Model) -> Any:
    return commands.getall(env.context, model, model.backend, resolved_params=env.params)


@ufunc.resolver(ParamBuilder, str)
def read(env: ParamBuilder, obj: str):
    new_name = obj
    if '/' not in obj:
        model_ = env.this
        if isinstance(model_, Model) and model_.external and model_.external.dataset:
            new_name = '/'.join([
                model_.external.dataset.name,
                obj,
            ])
    model = None
    if commands.has_model(env.context, env.manifest, new_name):
        model = commands.get_model(env.context, env.manifest, new_name)
    elif obj != new_name and commands.has_model(env.context, env.manifest, obj):
        model = commands.get_model(env.context, env.manifest, obj)
    if not model:
        raise ModelNotFound(model=obj)

    env.this = model
    return env.call("read", model)


@ufunc.resolver(ParamBuilder)
def read(env: ParamBuilder) -> Any:
    if isinstance(env.this, Model):
        return env.call("read", env.this)
    raise InvalidParamSource(param=env.target_param, source=env.this, given_type=type(env.this), expected_types=[type(Model)])


@ufunc.resolver(ParamBuilder, Iterator, Bind, name="getattr")
def getattr_(env: ParamBuilder, iterator: Iterator, bind: Bind):
    for item in iterator:
        yield from env.call("getattr", item, bind)


@ufunc.resolver(ParamBuilder, dict, Bind, name="getattr")
def getattr_(env: ParamBuilder, data: dict, bind: Bind):
    if bind.name in data:
        yield data[bind.name]
    else:
        raise KeyNotFound(env.this, key=bind.name, dict_keys=list(data.keys()))


@ufunc.resolver(ParamBuilder, NotAvailable, name="getattr")
def getattr_(env: ParamBuilder, _: NotAvailable):
    return env.this


# {'name': 'getattr', 'args': [{'name': 'loop', 'args': [{'name': 'read', 'args': []}], 'type': 'method'}, {'name': 'bind', 'args': ['more']}]}
# getattr [ loop(read()), bind(more) ]
#  1 -> stack = [1]
#  loop stack [1]
#  read(1) -> 2
#  pop(1) from stack
#  2 -> stack
#  loop until stack is empty


@ufunc.executor(ParamBuilder, NotAvailable)
def read(env: ParamBuilder, na: NotAvailable) -> Any:
    return env.this


@ufunc.resolver(ParamLoader, list)
def resolve_param(env: ParamLoader, params: list):
    for parameter in params:
        env.call("resolve_param", parameter)


@ufunc.resolver(ParamLoader, Param)
def resolve_param(env: ParamLoader, parameter: Param):
    for i, (source, prepare) in enumerate(zip(parameter.sources.copy(), parameter.formulas)):
        formula = None
        if isinstance(prepare, Expr):
            formula = Expr(prepare.name, *prepare.args, **prepare.kwargs)
        requires_model = env.call("contains_read", formula)
        if isinstance(source, str) and requires_model:
            new_name = source
            if env.dataset and '/' not in source:
                new_name = '/'.join([
                    env.dataset.name,
                    source,
                ])
            if commands.has_model(env.context, env.manifest, new_name):
                model = commands.get_model(env.context, env.manifest, new_name)
                parameter.sources[i] = model
            elif source != new_name and commands.has_model(env.context, env.manifest, source):
                model = commands.get_model(env.context, env.manifest, source)
                parameter.sources[i] = model
            else:
                raise ModelNotFound(model=source)

            env.call("validate_prepare", model, formula)


@ufunc.resolver(ParamLoader, Expr)
def contains_read(env: ParamLoader, expr: Expr):
    resolved, _ = expr.resolve(env)
    for arg in resolved:
        if isinstance(arg, bool) and arg is True:
            return True
    return False


@ufunc.resolver(ParamLoader, object)
def contains_read(env: ParamLoader, obj: object):
    return False


@ufunc.resolver(ParamLoader, Model, Expr)
def validate_prepare(env: ParamLoader, model: Model, expr: Expr):
    resolved, _ = expr.resolve(env)
    for arg in resolved:
        if isinstance(arg, str):
            if arg not in model.properties:
                raise PropertyNotFound(model, property=arg)
        elif isinstance(arg, Bind):
            if arg.name not in model.properties:
                raise PropertyNotFound(model, property=arg.name)


@ufunc.resolver(ParamLoader, object)
def read(env: ParamLoader, any_: object):
    return True


@ufunc.resolver(ParamLoader)
def read(env: ParamLoader):
    return True




from spinta import commands
from spinta.components import Property
from spinta.datasets.backends.sql.components import Sql
from spinta.exceptions import NoExternalName, PropertyNotFound
from spinta.types.datatype import DataType, Ref
from spinta.types.text.components import Text
import sqlalchemy as sa


@commands.get_column.register(Sql, Property)
def get_column(backend: Sql, prop: Property, **kwargs):
    return commands.get_column(backend, prop.dtype, **kwargs)


@commands.get_column.register(Sql, DataType)
def get_column(backend: Sql, dtype: DataType, table: sa.Table, **kwargs):
    prop = dtype.prop
    if prop.external is None or not prop.external.name:
        raise NoExternalName(prop)
    if prop.external.name not in table.c:
        raise PropertyNotFound(
            prop.model,
            property=prop.name,
            external=prop.external.name,
        )
    return table.c[prop.external.name]


@commands.get_column.register(Sql, Ref)
def get_column(backend: Sql, dtype: Ref, table: sa.Table, **kwargs):
    prop = dtype.prop
    if prop.external is None or not prop.external.name and not dtype.inherited:
        raise NoExternalName(prop)

    if prop.external.name:
        if prop.external.name not in table.c:
            raise PropertyNotFound(
                prop.model,
                property=prop.name,
                external=prop.external.name,
            )
        return table.c[prop.external.name]



# round_point("POINT(0.855556 5.100545)", 2) -> "POINT(0.86 5.1)"
def round_point(point: str, precision: int):
    if not '.' in point:
        return point

    prefix = 'POINT'
    if 'POINT ' in point:
        prefix = 'POINT '

    converted_point = point.replace(prefix, '')
    converted_point = converted_point.strip().removeprefix('(').removesuffix(')')
    split = converted_point.split(' ')
    if len(split) == 2:
        lat = float(split[0])
        lon = float(split[1])

        rounded_lat = round(lat, precision)
        rounded_lon = round(lon, precision)

        return f"{prefix}({rounded_lat} {rounded_lon})"
    return point


# round_url("'https://www.openstreetmap.org/?'
#     'mlat=54.68569111173754&'
#     'mlon=25.286688302053335'
#     '#map=19/54.68569111173754/25.286688302053335'", 2) ->
#     "'https://www.openstreetmap.org/?'
#     'mlat=54.69&'
#     'mlon=25.29'
#     '#map=19/54.69/25.29'"
def round_url(url: str, precision: int):
    converted_url = url.replace('https://www.openstreetmap.org/?', '')
    if not '.' in converted_url:
        return url

    split = converted_url.split('/')
    if len(split) == 3:
        lat = float(split[1])
        lon = float(split[2])

        rounded_lat = round(lat, precision)
        rounded_lon = round(lon, precision)

        return f"https://www.openstreetmap.org/?mlat={rounded_lat}&mlon={rounded_lon}#map=19/{rounded_lat}/{rounded_lon}"
    return url



from collections.abc import Callable

from typing import Optional, Any
from typing_extensions import Concatenate, ParamSpec

from spinta.components import Context

P = ParamSpec("P")
UpgradeFuncType = Callable[Concatenate[Context, P], Any]
UpgradeCheckFuncType = Callable[Concatenate[Context, P], bool]


class UpgradeComponent:
    def __init__(
        self,
        upgrade: UpgradeFuncType,
        check: Optional[UpgradeCheckFuncType] = None
    ):
        self.__upgrade_func = upgrade
        self.__check_func = check

    def upgrade(self, context: Context, **kwargs):
        self.__upgrade_func(context, **kwargs)

    def check(self, context: Context, **kwargs) -> bool:
        if self.__check_func is None:
            return True

        return self.__check_func(context, **kwargs)


UPGRADE_CLIENTS_SCRIPT = "clients"



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.ufuncs.result.components import PgResultBuilder
from spinta.components import Context


@commands.get_result_builder.register(Context, PostgreSQL)
def get_result_builder(
    context: Context,
    backend: PostgreSQL
):
    return PgResultBuilder(context)



def script_check_status_message(
    script_name: str,
    status: str
) -> str:
    return f"Script '{script_name}' upgrade check. Status: {status}"


def script_destructive_warning(
    script_name: str,
    message: str
) -> str:
    return f"WARNING (DESTRUCTIVE MODE). Script '{script_name}' will {message}."



from typing import List

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import json_has_key, get_root_attr, jsonb_keys, \
    MigratePostgresMeta, MigrateModelMeta
from spinta.backends.postgresql.helpers.migrate.name import name_changed, get_pg_removed_name
from spinta.components import Context
from spinta.types.datatype import String
from spinta.types.text.helpers import determine_langauge_for_text
from spinta.utils.nestedstruct import get_last_attr
from spinta.utils.schema import NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, String)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: String, model_meta: MigrateModelMeta = None, **kwargs):
    rename = meta.rename
    handler = meta.handler

    column: sa.Column = commands.prepare(context, backend, new.prop)
    old_name = rename.get_old_column_name(table.name, column.name, root_only=False)
    columns = old.copy()

    table_name = rename.get_table_name(table.name)

    json_column = None
    for col in old:
        if isinstance(col.type, JSONB):
            json_column = col
            break

    json_column_meta = None
    if json_column is not None and model_meta is not None:
        if json_column.name in model_meta.json_columns:
            json_column_meta = model_meta.json_columns[json_column.name]

    requires_removal = True

    # Check if column was renamed and if there already existed column of the new name
    # If it did, remove it
    if name_changed(column.name, old_name):
        if json_column is not None:
            key = get_last_attr(old_name)
            if json_column_meta and json_column_meta.keys:
                contains_key = key in json_column_meta.keys
            else:
                contains_key = json_has_key(backend, json_column, table, key)
            requires_removal = contains_key

            if not contains_key:
                columns.remove(json_column)

        if requires_removal:
            for col in old:
                if not name_changed(col.name, column.name):
                    commands.migrate(context, backend, meta, table, col, NA, **kwargs)
                    columns.remove(col)
                    break

    for col in columns.copy():
        if col.name != column.name and not isinstance(col.type, JSONB):
            name = rename.get_old_column_name(table.name, col.name)
            if name != col.name:
                name = get_root_attr(name)
                if json_column is not None and name == json_column.name:
                    columns.remove(col)

    if len(columns) <= 1:
        col = columns[0] if len(columns) == 1 else NA

        # Check if it's text -> string, or just normal string migration
        if col != NA and isinstance(col.type, JSONB):
            old_name = rename.get_old_column_name(table.name, column.name, root_only=False)
            key = get_last_attr(old_name)

            # No key was extracted
            cast_json_to_string = key == old_name
            if cast_json_to_string:
                default_langs = context.get('config').languages
                if json_column_meta and json_column_meta.keys:
                    all_keys = json_column_meta.keys
                else:
                    all_keys = jsonb_keys(backend, column, table)
                key = determine_langauge_for_text(all_keys, [], default_langs)
                json_column_meta.cast_to = (column, key)
            else:
                commands.migrate(context, backend, meta, table, NA, column, **kwargs)
                handler.add_action(
                    ma.TransferJSONDataMigrationAction(table_name, col, columns=[
                        (key, column)
                    ])
                )
                renamed_key = get_pg_removed_name(key)
                if json_column_meta is None:
                    if json_has_key(backend, col, table, renamed_key):
                        handler.add_action(
                            ma.RemoveJSONAttributeMigrationAction(table_name, col, renamed_key)
                        )
                    handler.add_action(
                        ma.RenameJSONAttributeMigrationAction(table_name, col, key, renamed_key)
                    )
                else:
                    json_column_meta.add_new_key(key, renamed_key)
        else:
            commands.migrate(context, backend, meta, table, col, column, **kwargs)
    else:
        raise Exception("String cannot migrate to multiple columns")


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, String)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: String, **kwargs):
    commands.migrate(context, backend, meta, table, [old], new, **kwargs)



import contextlib
from typing import Optional, Dict, List, Final, Literal

from spinta.manifests.components import Manifest
import sqlalchemy as sa


class InternalSQLManifest(Manifest):
    type = 'internal'
    path: Optional[str] = None
    engine: sa.engine.Engine = None
    table = None

    dynamic = True

    @staticmethod
    def detect_from_path(path: str) -> bool:
        try:
            url = sa.engine.make_url(path)
            if not url:
                return False
            url.get_dialect()
            engine = sa.create_engine(url)
            inspector = sa.inspect(engine)
            return inspector.has_table('_manifest')
        except sa.exc.SQLAlchemyError:
            return False

    @contextlib.contextmanager
    def transaction(self):
        with self.engine.begin() as connection:
            yield Transaction(connection)


class Transaction:
    id: str
    errors: int

    def __init__(self, connection):
        self.connection = connection


INDEX: Final = 'index'
ID: Final = 'id'
PARENT: Final = 'parent'
DEPTH: Final = 'depth'
PATH: Final = 'path'
MPATH: Final = 'mpath'
DIM: Final = 'dim'
NAME: Final = 'name'
TYPE: Final = 'type'
REF: Final = 'ref'
SOURCE: Final = 'source'
PREPARE: Final = 'prepare'
LEVEL: Final = 'level'
ACCESS: Final = 'access'
URI: Final = 'uri'
TITLE: Final = 'title'
DESCRIPTION: Final = 'description'
InternalManifestColumn = Literal[
    'index',
    'id',
    'parent',
    'depth',
    'path',
    'mpath',
    'dim',
    'name',
    'type',
    'ref',
    'source',
    'prepare',
    'level',
    'access',
    'uri',
    'title',
    'description',
]
INTERNAL_MANIFEST_COLUMNS: List[InternalManifestColumn] = [
    INDEX,
    ID,
    PARENT,
    DEPTH,
    PATH,
    MPATH,
    DIM,
    NAME,
    TYPE,
    REF,
    SOURCE,
    PREPARE,
    LEVEL,
    ACCESS,
    URI,
    TITLE,
    DESCRIPTION,
]

InternalManifestRow = Dict[InternalManifestColumn, str]



from typing import Union, Tuple

import sqlalchemy as sa


def extract_dialect(engine: sa.engine.Engine):
    if engine is not None:
        return engine.dialect.name
    return _DEFAULT_DIALECT_KEY


def does_dialect_match(src_dialect: str, target_dialect: Union[str, Tuple[str]]):
    if not target_dialect or not src_dialect:
        return False

    if isinstance(target_dialect, str):
        return target_dialect in src_dialect
    elif isinstance(target_dialect, tuple):
        for dialect in target_dialect:
            if dialect in src_dialect:
                return True

    return False


def _nulls_asc(column: sa.Column):
    return sa.sql.expression.nullslast(column.asc())


# Reason for column == None is NULLS LAST, because
# if it's NULL it will return 1 and if it's not NULL it will return 0
# when ordering 0 takes priority over 1, so then it will sort first values that are not NULL
def _default_asc(column: sa.Column):
    return sa.case(
        [
            (
                column == None,
                sa.literal_column('1', type_=sa.Integer)
            )
        ],
        else_=sa.literal_column('0', type_=sa.Integer)
    ), column.asc()


def _nulls_desc(column: sa.Column):
    return sa.sql.expression.nullsfirst(column.desc())


# Reason for column != None is NULLS FIRST, because
# if it's NULL it will return 0 and if it's not NULL it will return 1
# when ordering 0 takes priority over 1, so then it will sort first values that are NULL
def _default_desc(column: sa.Column):
    return sa.case(
        [
            (
                column != None,
                sa.literal_column('1', type_=sa.Integer)
            )
        ],
        else_=sa.literal_column('0', type_=sa.Integer)
    ), column.desc()


_DEFAULT_DIALECT_KEY = ""


_DESC_DIALECT_MAPPER = {
    ("postgresql", "oracle", "sqlite"): _nulls_desc,
    _DEFAULT_DIALECT_KEY: _default_desc
}
_ASC_DIALECT_MAPPER = {
    ("postgresql", "oracle", "sqlite"): _nulls_asc,
    _DEFAULT_DIALECT_KEY: _default_asc
}


def _dialect_specific_function(engine: sa.engine.Engine, dialect_function_mapper: dict, **kwargs):
    dialect = extract_dialect(engine)
    for key, func in dialect_function_mapper.items():
        if key != _DEFAULT_DIALECT_KEY and does_dialect_match(dialect, key):
            return func(**kwargs)

    return dialect_function_mapper[_DEFAULT_DIALECT_KEY](**kwargs)


def dialect_specific_desc(engine: sa.engine.Engine, column: sa.Column):
    return _dialect_specific_function(
        engine=engine,
        dialect_function_mapper=_DESC_DIALECT_MAPPER,
        column=column
    )


def dialect_specific_asc(engine: sa.engine.Engine, column: sa.Column):
    return _dialect_specific_function(
        engine=engine,
        dialect_function_mapper=_ASC_DIALECT_MAPPER,
        column=column
    )
